import dataclasses
from typing import Callable, Tuple, Any, List, Optional, Dict
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.quantized.dynamic as nnqd
from intel_extension_for_pytorch.nn.functional import interaction
import intel_extension_for_pytorch._C as core


functions_supported_by_quantization =set([
    torch.Tensor.add,
    torch.add,
    torch.Tensor.relu,
    #torch.Tensor.sigmoid,  # TODO
    torch.flatten,
    torch.Tensor.flatten,
    F.adaptive_avg_pool2d,
    F.adaptive_avg_pool3d,
    F.avg_pool2d,
    F.avg_pool3d,
    F.max_pool2d,
    F.max_pool3d,
    F.conv2d,
    F.conv3d,
    torch.conv2d,
    torch.conv3d,
    #F.conv_transpose2d,   #TODO
    #F.conv_transpose3d,   #TODO
    #torch.conv_transpose2d,  #TODO
    #torch.conv_transpose3d,  #TODO
    torch.relu,
    F.relu,
    #torch.sigmoid,  # TODO
    #F.sigmoid,  # TODO
    #F.gelu, # TODO
    F.linear,
    torch._C._nn.linear,
    torch.matmul,
    torch.Tensor.matmul,
    F.embedding_bag,
    torch.embedding_bag,
    ])

# ipex customer function
functions_supported_by_quantization_ipex = set([
    interaction,
    torch.ops.torch_ipex.interaction_forward,
    ])

module_types_supported_by_quantization = set([
    torch.nn.Conv2d,
    torch.nn.Conv3d,
    #torch.nn.ConvTranspose2d,
    #torch.nn.ConvTranspose3d,
    torch.nn.Linear,
    torch.nn.MaxPool2d,
    torch.nn.MaxPool3d,
    torch.nn.AvgPool2d,
    torch.nn.AvgPool3d,
    torch.nn.AdaptiveAvgPool2d,
    torch.nn.AdaptiveAvgPool3d,
    torch.nn.ReLU,
    #torch.nn.Sigmoid,  # TODO
    #torch.nn.GELU,     # TODO
    torch.nn.EmbeddingBag,
    torch.nn.Flatten,
    torch.nn.LSTM,
    # dynamic quantization module
    nnqd.Linear,
    nnqd.LSTM,
    ])

may_inplace_module = set([
    torch.nn.ReLU,
    ])


a_related_to_b = (
    (str(torch.add), str(torch.Tensor.add)),
    (str(torch.Tensor.add), str(torch.add)),
    (str(torch.nn.Linear), str(nnqd.Linear)),
    (str(nnqd.Linear), str(torch.nn.Linear)),
    (str(torch.nn.LSTM), str(nnqd.LSTM)),
    (str(nnqd.LSTM), str(torch.nn.LSTM)),
)

conv_linear_ops = [
    #F.conv1d, # it will be enabled at next step.
    str(F.conv2d),
    str(F.conv3d),
    str(torch.conv2d),
    str(torch.conv3d),
    #str(F.conv_transpose2d),
    #str(F.conv_transpose3d),
    #str(torch.conv_transpose2d),
    #str(torch.conv_transpose3d),
    str(F.linear), 
    str(torch._C._nn.linear),
    ]

conv_linear_modules = [
    #str(torch.nn.Conv1d) # it will be enabled at next step.
    str(torch.nn.Conv2d),
    str(torch.nn.Conv3d),
    #str(torch.nn.ConvTranspose2d),
    #str(torch.nn.ConvTranspose3d),
    str(torch.nn.Linear),
    ]

embedding_op = [
    str(F.embedding_bag),
    str(torch.embedding_bag),
    ]

def op_needs_quantization(op: Callable) -> bool:
    if op in functions_supported_by_quantization or op in functions_supported_by_quantization_ipex:
        return True
    elif type(op) in module_types_supported_by_quantization:
        if op in may_inplace_module and op.inplace:
            return False
        return True
    else:
        return False

def ops_are_related(
    cur_op: Callable,
    expected_op_type: str,
    type_is_module: bool,
) -> bool:
    r"""
    This function is to check whether the cur_op is align with the saved op_type, which make sure
    the model doesn't have dynamic workflow.
    """
    if type_is_module:
        cur_op = type(cur_op)
    return str(cur_op) == expected_op_type or \
        (str(cur_op), expected_op_type) in a_related_to_b

def _raise_obs_not_found_error(func):
    raise RuntimeError(
        f'Encountered arithmetic operation {torch.typename(func)} but we have '
        f'encountered fewer arithmetic operations in previous calibration runs. '
        f'This likely indicates that the program contains dynamic control flow. '
        f' Quantization is not defined over dynamic control flow!')

def _raise_obs_op_mismatch(func, prev_op):
    raise RuntimeError(
        f'Encountered arithmetic operation {torch.typename(func)} but previously '
        f'recorded operation was {prev_op}!. This likely indicates '
        f'that the program contains dynamic control flow. Quantization is not '
        f'defined over dynamic control flow!')

@dataclasses.dataclass
class QTensorInfo:
    id: int  # tensor ID
    orig_dtype: torch.dtype  # dtype seen while tracing with example input
    inf_dtype: torch.dtype  # dtype at inference

@dataclasses.dataclass
class SeenQOpInfo:
    idx: int
    # Python type of the seen op. For modules, this is str(type(mod)). For
    # functions, this is the target function(str).
    type: str
    # True if the type is a module, False otherwise (for functions/methods).
    type_is_module: bool
    # Note: FQN refers to the current module for modules and to the parent
    # module for functions
    fqn: str
    # Information about the input tensors
    # Non-tensor inputs are represented with None.
    input_tensor_infos: List[Optional[QTensorInfo]]
    # We use input_tensor_infos's inf_dtype to check whether we need add fake quant
    # at convert step, but sometimes, the QTensorInfo's infor may used by many 
    # operators, and one operator may set QTensorInfo' inf dtype to fp32, which hope
    # use fp32 kernel, but the cur op hope use low-precison op, so we introduce this flag
    # to fix the multi-use case: if input_tensor_force_inf_dtype has low-precison, we will
    # ignore the related QTensorInfo's inf dtype even QTensorInfo's inf dtype is fp32 dtype.
    # Note: the inint value of the QTensorInfo's  is orig dtype.
    input_tensor_force_inf_dtype: List[Optional[torch.dtype]]
    # Information about the output tensors
    # Non-tensor outputs are represented with None.
    output_tensor_infos: List[QTensorInfo]
    # Some operator only support INT8->INT8, if post operator is non-quantized op,
    # the output_tensor_infos's inf dtype always same as orig dtype, we can set the output_tensor_infos's
    # inf dtype to int8, and do a check whether add fake quant after output accoreding to the inf dtype,
    # but if the post operator is quantized op, we will add two fake quant if we only check the inf dtype.
    # so we introduce insert_fake_quant_after_output to fix this issue: if insert_fake_quant_after_output is true,
    # and the the inf dtype is int8, we will add fake quant after the output, otherwise, we will not insert fake quant
    # after the output(if inf dtype is int8, but insert_fake_quant_after_output is False, the post op will insert
    # fake quant, if inf dtype is not int8, the output hopes a orig dtype, we don't need to add fake quant).
    # Note: the init value of the insert_fake_quant_after_output's is False.
    # Our Quant param binding algorithm (binding info used to decide whether to add q/dq at runtime) is that:
    # 1. Bind input tensors by default for all quantized ops.
    # 2. Bind output tensor if any of downstream ops is not quantized.
    insert_fake_quant_after_outputs: List[Optional[bool]]
    weight_tensor_infos: List[Optional[QTensorInfo]]
    qconfig:  torch.ao.quantization.QConfig

    def __repr__(self) -> str:
        s = f"(type): {self.type}\n"
        s += f"     (fqn): {self.fqn}\n"
        s += f"     (input_tensor_infos): {self.input_tensor_infos}\n"
        s += f"     (input_tensor_force_inf_dtype): {self.input_tensor_force_inf_dtype}\n"
        s += f"     (output_tensor_infos): {self.output_tensor_infos}\n"
        s += f"     (insert_fake_quant_after_outputs): {self.insert_fake_quant_after_outputs}\n"
        s += f"     (weight_tensor_infos): {self.weight_tensor_infos}\n"
        s += f"     (qconfig): {self.qconfig}"
        return s

@dataclasses.dataclass
class SeenNonQOpInfo:
    # Python type of the seen op. For modules, this is str(type(mod)). For
    # functions, this is the target function.
    type: str
    # Note: FQN refers to the current module for modules and to the parent
    # module for functions
    fqn: str
    # Information about the input tensors
    # Non-tensor inputs are represented with None.
    input_tensor_infos: List[Optional[QTensorInfo]]
    # Information about the output tensors
    # Non-tensor outputs are represented with None.
    output_tensor_infos: List[QTensorInfo]

def get_input_observed_arg_idxs(
    op_type: str,
    op_type_is_module: bool,
) -> Optional[List[int]]:
    if op_type_is_module and op_type != str(torch.nn.EmbeddingBag):
        # TODO(future PR): handle RNNs
        return [0]
    elif op_type in conv_linear_ops:
        return [0, 1]
    elif op_type in embedding_op:
        return [1]
    # None means "observe all Tensor args"
    return None

def get_weight_arg_idx(op: str) -> Optional[int]:
    if op in conv_linear_ops:
        return 1
    return None

def iterate_and_apply(
    args: Any,
    flattened_tensor_infos: List[Optional[QTensorInfo]],
    func: Callable,
    flattened_tensor_infos_idx=None
) -> Any:
    """
    Inputs:
      `args`: arguments to a function, may contain nested types, for example:
        ([torch.Tensor, torch.Tensor], int, (int, int))
      `flattened_tensor_infos`: tensor information containers for each tensor
        in `args`, flattened, for example corresponding with above:
        ({...}, {...}, None, None, None)
      `func`: function to apply to each tensor in `args` to create `new_args`
    Returns `new_args`, where each tensor has been transformed by `func`.
    """
    if flattened_tensor_infos_idx is None:
        flattened_tensor_infos_idx = [0]

    if isinstance(args, tuple):
        new_args = []
        for arg in args:
            new_arg = iterate_and_apply(
                arg, flattened_tensor_infos, func, flattened_tensor_infos_idx)
            new_args.append(new_arg)
        return tuple(new_args)
    elif isinstance(args, list):
        for idx in range(len(args)):
            new_arg = iterate_and_apply(
                args[idx], flattened_tensor_infos, func, flattened_tensor_infos_idx)
            args[idx] = new_arg
        return args
    else:
        # individual element
        cur_flattened_tensor_info = \
            flattened_tensor_infos[flattened_tensor_infos_idx[0]]
        flattened_tensor_infos_idx[0] += 1

        if cur_flattened_tensor_info is not None:
            return func(args, cur_flattened_tensor_info)
        else:
            return args

def iterate_and_apply_convert(
    args: Any,
    quant_infos: List[Optional[Tuple[float, int, torch.dtype]]],
    quant_or_dequant_needed: List[bool],
    op: Callable,
    flattened_tensor_infos_idx=None
) -> Any:
    """
    Inputs:
      `args`: arguments to a function, may contain nested types, for example:
        ([torch.Tensor, torch.Tensor], int, (int, int))
      `quant_infos`: tensor information containers for each tensor
        in `args`, flattened, for example corresponding with above:
        ({...}, {...}, None, None, None)
       `quant_or_dequant_needed`: tensor information about whether do quantization
        containers for each tensorin `args`,
      `op`: cur quantizable op
    Returns `new_args`, where each tensor has been transformed by `func`.
    """
    
    if flattened_tensor_infos_idx is None:
        flattened_tensor_infos_idx = [0]
    if isinstance(args, tuple):
        new_args = []
        for arg in args:
            new_arg = iterate_and_apply_convert(
                arg, quant_infos, quant_or_dequant_needed, op, flattened_tensor_infos_idx)
            new_args.append(new_arg)
        return tuple(new_args)
    elif isinstance(args, list):
        new_args = []
        for arg in args:
            new_arg = iterate_and_apply_convert(
                arg, quant_infos, quant_or_dequant_needed, op, flattened_tensor_infos_idx)
            new_args.append(new_arg)
        return new_args
    else:
        # individual element
        cur_quant_infos = \
            quant_infos[flattened_tensor_infos_idx[0]]
        cur_quant_or_dequant_needed = \
            quant_or_dequant_needed[flattened_tensor_infos_idx[0]]
        if cur_quant_infos is not None and cur_quant_or_dequant_needed and isinstance(args, torch.Tensor):
            scale, zp, dtype = cur_quant_infos
            # For F.Linear, F.conv, the weight's may use per_channel.
            if str(op) in conv_linear_ops and get_weight_arg_idx(str(op)) == flattened_tensor_infos_idx[0] and isinstance(scale, torch.Tensor) and scale.numel() > 1:
                ch_axis = 0
                # conv_transpose's weight is iohw or iodhw
                if str(op) in [str(F.conv_transpose2d), str(torch.conv_transpose2d), str(F.conv_transpose3d), str(torch.conv_transpose3d)]:
                    ch_axis = 1
                # core.get_autocast_dtype() will be removed after fully use pytorch autocast
                if torch.is_autocast_cpu_enabled() and core.get_autocast_dtype() == torch.bfloat16:
                    # do autocast in Python side
                    if args.dtype == torch.float32:
                        args = args.to(torch.bfloat16)
                    args = args.to(torch.float32)
                    args = torch.quantize_per_channel(args, scale, zp, ch_axis, dtype)
                    args = args.dequantize()
                    args = arg.to(torch.bfloat16)
                else:
                    args = torch.quantize_per_channel(args, scale, zp, ch_axis, dtype)
                    args = args.dequantize()
            else:
                # white list, conv, linear, matmul, we always convert it's input to bflat16 firstly, and then inser q+dq
                if str(op) in conv_linear_ops + [str(torch.matmul), str(torch.Tensor.matmul)] + embedding_op or str(type(op)) in conv_linear_modules:
                    if torch.is_autocast_cpu_enabled() and core.get_autocast_dtype() == torch.bfloat16:
                        if args.dtype == torch.float32:
                            args = args.to(torch.bfloat16)
                        args = args.to(torch.float32)
                        args = torch.quantize_per_tensor(args, scale.item(), zp.item(), dtype)
                        args = args.dequantize()
                        args = args.to(torch.bfloat16)
                    else:
                        args = torch.quantize_per_tensor(args, scale.item(), zp.item(), dtype)
                        args = args.dequantize()
                else:
                    # fall through
                    args_is_bfloat16 = False
                    if args.dtype == torch.bfloat16:
                        args_is_bfloat16 = True
                        args = args.to(torch.float32)
                    args = torch.quantize_per_tensor(args, scale.item(), zp.item(), dtype)
                    args = args.dequantize()
                    if args_is_bfloat16:
                        args = args.to(torch.bfloat16)
        flattened_tensor_infos_idx[0] += 1
        return args

def get_input_args_quant_dequant_info(
    seen_q_op_info: SeenQOpInfo,
    tensor_id_to_scale_zp: Dict[int, Tuple[torch.Tensor, torch.Tensor]],
) -> Tuple[List[Optional[Tuple[float, int, torch.dtype]]], List[bool], bool]:
    """
    Returns a list of information about the tensor inputs to the current op.
    Quant list:
    For each tensor input:
    * if the tensor input needs a quant, the list will contain
      (scale, zero_point)
    * if the tensor input does not need a quant, the list will contain None
    """
    quant_infos: List[Optional[Tuple[float, int, torch.dtype]]] = []
    quantized_dtype = [torch.quint8, torch.qint8]
    any_arg_quant_or_dequant_needed = []
    if len(seen_q_op_info.input_tensor_infos) > 0:
        for i, input_arg in enumerate(seen_q_op_info.input_tensor_infos):
            if input_arg is not None:
                if input_arg.id in tensor_id_to_scale_zp:
                    tensor_id = input_arg.id
                    inf_dtype = input_arg.inf_dtype
                    # force_inf_dtype always should be same as input_arg.inf_dtype, but some time,
                    # the input arg may be used by many other operators, and it may have been
                    # changed by other operators, so for cur op, twe check whether input_arg.inf_dtype
                    # is same as the origin force_inf_dtype, if not same use force_inf_dtype as new
                    # inf dtype, if same, we can say the input_arg.inf_dtype is not changed or the cur op
                    # changed input_arg.inf_dtype and force_inf_dtype at get default recipe step.
                    if seen_q_op_info.input_tensor_force_inf_dtype[i] != input_arg.inf_dtype:
                        inf_dtype = seen_q_op_info.input_tensor_force_inf_dtype[i]
                
                    scale, zp = tensor_id_to_scale_zp[tensor_id]
                    quant_infos.append((scale, zp, inf_dtype))  # type: ignore[arg-type]
                    # only support float to int8.
                    if input_arg.orig_dtype == torch.float32 and inf_dtype in quantized_dtype:
                        any_arg_quant_or_dequant_needed.append(True)
                    else:
                        any_arg_quant_or_dequant_needed.append(False)
                else:
                     quant_infos.append(None)
                     any_arg_quant_or_dequant_needed.append(False)
            else:
                quant_infos.append(None)
                any_arg_quant_or_dequant_needed.append(None)
    return quant_infos, any_arg_quant_or_dequant_needed

def get_weight_args_quant_dequant_info(
    seen_q_op_info: SeenQOpInfo,
    weight_tensor_id_to_scale_zp: Dict[str, Tuple[torch.Tensor, torch.Tensor]],
) -> Tuple[List[Optional[Tuple[float, int, torch.dtype]]], List[bool], bool]:
    """
    Returns a list of information about the tensor inputs to the current op.
    """
    quant_infos: List[Optional[Tuple[float, int, torch.dtype]]] = []
    any_arg_quant_or_dequant_needed = []
    for _, input_arg in enumerate(seen_q_op_info.weight_tensor_infos):
        if input_arg is not None:
            tensor_id = str(seen_q_op_info.idx) + "_" + str(input_arg.id)
            if tensor_id in weight_tensor_id_to_scale_zp:
                scale, zp = weight_tensor_id_to_scale_zp[tensor_id]
                output_dtype = input_arg.inf_dtype
                quant_infos.append((scale, zp, output_dtype))  # type: ignore[arg-type]
                if input_arg.orig_dtype == torch.float32 and input_arg.inf_dtype in [torch.quint8,  torch.qint8]:
                    any_arg_quant_or_dequant_needed.append(True)
                else:
                    any_arg_quant_or_dequant_needed.append(False)
            else:
                quant_infos.append(None)
                any_arg_quant_or_dequant_needed.append(False)
        else:
            quant_infos.append(None)
            any_arg_quant_or_dequant_needed.append(None)
    return quant_infos, any_arg_quant_or_dequant_needed



import torch
import torch.nn as nn
from typing import Union, Optional
import intel_extension_for_pytorch._C as core
from .cpupool import CPUPool
from .task import Task
import copy
import warnings

class MultiStreamModuleHint(object):
    r"""
    MultiStreamModuleHint is a hint to MultiStreamModule about how to split the inputs
    or concat the output. Each argument should be None, with type of int or a container
    which containes int or None such as: (0, None, ...) or [0, None, ...]. If the argument
    is None, it means this argument will not be split or concat. If the argument is with
    type int, its value means along which dim this argument will be split or concat.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        intel_extension_for_pytorch.cpu.runtime.MultiStreamModuleHint: Generated
        intel_extension_for_pytorch.cpu.runtime.MultiStreamModuleHint object.

    :meta public:
    """

    def __init__(self, *args, **kwargs):
        self.args = list(args)
        self.kwargs = kwargs
        self.args_len = args.__len__()
        self.kwargs_len = kwargs.__len__()

default_multi_stream_module_split_hint = MultiStreamModuleHint(0)
default_multi_stream_module_concat_hint = MultiStreamModuleHint(0)

def get_default_num_streams(cpu_pool):
    # One core per stream usually brings better overall throughput than other configurations.
    # Therefore, we heuristically make one core per stream the default here.
    return cpu_pool.core_ids.__len__()

class MultiStreamModule(nn.Module):
    r"""
    MultiStreamModule supports inference with multi-stream throughput mode.

    If the number of cores inside ``cpu_pool`` is divisible by ``num_streams``,
    the cores will be allocated equally to each stream. If the number of cores
    inside ``cpu_pool`` is not divisible by ``num_streams`` with remainder N,
    one extra core will be allocated to the first N streams. We suggest to set
    the ``num_streams`` as divisor of core number inside ``cpu_pool``.

    If the inputs' batchsize is larger than and divisible by ``num_streams``,
    the batchsize will be allocated equally to each stream. If batchsize is not
    divisible by ``num_streams`` with remainder N, one extra piece will be
    allocated to the first N streams. If the inputs' batchsize is less than
    ``num_streams``, only the first batchsize's streams are used with mini batch
    as one. We suggest to set inputs' batchsize larger than and divisible by
    ``num_streams``. If you don't want to tune the num of streams and leave it
    as "AUTO", we suggest to set inputs' batchsize larger than and divisible by
    number of cores.

    Args:
        model (torch.jit.ScriptModule or torch.nn.Module): The input model.
        num_streams (Union[int, str]): Number of instances (int) or "AUTO" (str). "AUTO" means the stream number
            will be selected automatically. Although "AUTO" usually provides a
            reasonable performance, it may still not be optimal for some cases which
            means manual tuning for number of streams is needed for this case.
        cpu_pool (intel_extension_for_pytorch.cpu.runtime.CPUPool): An
            intel_extension_for_pytorch.cpu.runtime.CPUPool object, contains
            all CPU cores used to run multi-stream inference.
        concat_output (bool): A flag indicates whether the output of each
            stream will be concatenated or not. The default value is True. Note:
            if the output of each stream can't be concatenated, set this flag to
            false to get the raw output (a list of each stream's output).
        input_split_hint (MultiStreamModuleHint): Hint to MultiStreamModule about
            how to split the inputs.
        output_concat_hint (MultiStreamModuleHint): Hint to MultiStreamModule about
            how to concat the outputs.

    Returns:
        intel_extension_for_pytorch.cpu.runtime.MultiStreamModule: Generated
        intel_extension_for_pytorch.cpu.runtime.MultiStreamModule object.

    :meta public:
    """
    def __init__(self,
                model,
                num_streams: Union[int, str] = "AUTO",
                cpu_pool: CPUPool = CPUPool(node_id=0),
                concat_output: bool = True,
                input_split_hint: MultiStreamModuleHint = default_multi_stream_module_split_hint,
                output_concat_hint: MultiStreamModuleHint = default_multi_stream_module_concat_hint):
        super(MultiStreamModule, self).__init__()
        assert type(cpu_pool) is CPUPool, "Input of cpu_pool must be provided with type of ipex.cpu.runtime.CPUPool"
        if not isinstance(model, torch.jit.ScriptModule):
            warnings.warn("Creating MultiStreamModule on an nn.Module. This can be slow due "
            "to Python Global Interpreter Lock (GIL). Suggest to use JIT ScriptModule for better performance.")
        self.core_list = cpu_pool.core_ids
        if isinstance(num_streams, str):
            # For str input of num_streams, it must be "auto"
            if num_streams.upper() == "AUTO":
                self.num_streams = get_default_num_streams(cpu_pool) # The default selected value when auto selection is on.
            else:
                assert False, "Input of num_streams must be Number of instances or string \"AUTO\""
        else:
            assert isinstance(num_streams, int), "Input of num_streams must be Number of instances or string \"auto\""
            self.num_streams = num_streams
        if self.num_streams == 1:
            # Sync execution path if num_stream is 1.
            self.model = model
        else:
            self.cores_per_instance = self.core_list.__len__() // self.num_streams
            num_stream_allocated_extra_core = self.core_list.__len__() % self.num_streams
            self.tasks = []
            start_core_list_idx = 0
            end_core_list_idx = 0
            for j in range(self.num_streams):
                if j < num_stream_allocated_extra_core:
                    # If the core number is not divisible by stream number,
                    # the remainder streams(num_stream_allocated_extra_core) will be allocated one extra core.
                    end_core_list_idx += (self.cores_per_instance + 1)
                else:
                    end_core_list_idx += self.cores_per_instance
                self.tasks.append(Task(model, CPUPool(self.core_list[start_core_list_idx:end_core_list_idx])))
                start_core_list_idx = end_core_list_idx
        self.concat_output = concat_output
        self.input_split_hint = input_split_hint
        self.output_concat_hint = output_concat_hint

        # Deep copy the input structure for each stream based on input_split_hint.
        # Each streams_input will be recursively visited and set to the split value in place.
        self.args_streams_input = []
        self.kwargs_streams_input = []
        for _ in range(self.num_streams):
            self.args_streams_input.append(copy.deepcopy(self.input_split_hint.args))
            self.kwargs_streams_input.append(copy.deepcopy(self.input_split_hint.kwargs))

        # Deep copy the output structure based on output_concat_hint.
        # self.output will be recursively visited and set to the concat value in place.
        self.output = copy.deepcopy(self.output_concat_hint)

        # Init status needed for forward
        self.reset_forward_status()

    def reset_forward_status(self):
        # Since the input batchsize for each forward invoking may change
        # Need to reset the status for each forward invoking
        #   * split_size: will be calculated by input batch size and num_streams.
        #   * used_num_streams: is the num_streams actually used by this forward invoking.
        #       It may less than self.num_streams when bs is less than self.num_streams.
        #   * current_split_start_idx: used to record the split start idx for current stream.
        #   * current_split_end_idx: used to record the split end idx for current stream.
        self.split_size = None
        self.used_num_streams = self.num_streams
        self.current_split_start_idx = 0
        self.current_split_end_idx = 0

    def update_split_idx(self, stream_id):
        # Set current_split_start_idx to last current_split_end_idx
        self.current_split_start_idx = self.current_split_end_idx
        # Calculate current_split_end_idx to new value
        if stream_id < self.instance_need_extra_input:
            # Tail case, when the input image size larger than num_streams and not divisible,
            # the first remainder streams will have (mini_batch + 1) input size.
            self.current_split_end_idx = self.current_split_end_idx + (self.batch_per_instance + 1)
        else:
            # Input image size divisible of num_streams or input image size less than num_streams.
            self.current_split_end_idx = self.current_split_end_idx + self.batch_per_instance

    def init_forward_status(self, split_size, stream_id):
        # This function should be invoke only once at each forward
        self.split_size = split_size
        # Ensure each instance has input offload
        self.batch_per_instance = self.split_size // self.num_streams
        if self.batch_per_instance >= 1:
            # The input batchsize larger or equal to num_streams.
            self.used_num_streams = self.num_streams
            # If input batchsize larger than num_streams and not divisible,
            # the first remainder streams will have (mini_batch + 1) input size.
            self.instance_need_extra_input = self.split_size % self.num_streams
        else:
            # The input batchsize less than num_streams,
            # only the first batchsize stream will have mini_batch(1) input.
            self.batch_per_instance = 1
            self.used_num_streams = self.split_size
            self.instance_need_extra_input = 0
        self.update_split_idx(stream_id)

    def _do_get_input_for_each_stream(self, hint_object, input_object, stream_input_object, idx_or_key, stream_id):
        #* hint_object: input hint to tell whether we need to split corresponding
        #       input_object at current position.
        #* input_object: raw input used to split and generate stream_input_object
        #       at current position.
        #* stream_input_object: Which will generated in place as the input for
        #       current stream (marked by stream_id)
        #* idx_or_key: idx (for list/tuple) and key (for dict) used for recursive
        #       visit of hint_object/input_object/stream_input_object.
        #* stream_id: the stream we are visiting now.
        type_arg = type(hint_object[idx_or_key])
        if type_arg in [list]:
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_get_input_for_each_stream(hint_object[idx_or_key], input_object[idx_or_key], stream_input_object[idx_or_key], i, stream_id)
        if type_arg in [tuple]:
            # Tuple doesn't support item change in place
            # So we change it to list for next recursion and change it back to tuple.
            temp = list(stream_input_object[idx_or_key])
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_get_input_for_each_stream(hint_object[idx_or_key], input_object[idx_or_key], temp, i, stream_id)
            stream_input_object[idx_or_key] = tuple(temp)
        elif type_arg in [dict]:
            for key in hint_object[idx_or_key]:
                self._do_get_input_for_each_stream(hint_object[idx_or_key], input_object[idx_or_key], stream_input_object[idx_or_key], key, stream_id)
        elif (type_arg is int) or (hint_object[idx_or_key] is None):
            if hint_object[idx_or_key] is not None:
                # If user tells us to split in this object,
                if self.split_size is None:
                    # Init the input status for each stream here
                    # Here the stream_id must be 0
                    self.init_forward_status(input_object[idx_or_key].size(hint_object[idx_or_key]), stream_id)
                # Get the split input for each stream
                # Here we assume split along the outside dim, otherwise memory copy happens and obviously hurt multi stream module's performance.
                if hint_object[idx_or_key] == 0:
                    # Split along dim 0, the slice will not create new tensor
                    stream_input_object[idx_or_key] = input_object[idx_or_key][self.current_split_start_idx:self.current_split_end_idx]
                else:
                    # Otherwise, we use torch.narrow
                    length = self.current_split_end_idx - self.current_split_start_idx
                    stream_input_object[idx_or_key] = input_object[idx_or_key].narrow(hint_object[idx_or_key], self.current_split_start_idx, length)
            else:
                # This object shouldn't be split, just set it as each stream's input
                stream_input_object[idx_or_key] = input_object[idx_or_key]
        else:
            assert False, "Generate stream input failed, unsupport input hint type of:{}".format(type_arg)
        return None

    def _get_input_for_each_stream(self, multi_stream_module_split_hint, *args, **kwargs):
        # recursive once to init:
        #   1. Decide the actual self.used_num_streams (it may less than number stream when input bs is small)
        #   2. Init the current_split_start_idx and current_split_end_idx for inputs split
        #   3. Decide the actual input for stream_id 0
        for i in range(multi_stream_module_split_hint.args_len):
            self._do_get_input_for_each_stream(hint_object = multi_stream_module_split_hint.args,
                                            input_object = args,
                                            stream_input_object = self.args_streams_input[0],
                                            idx_or_key = i,
                                            stream_id = 0)
        for key in multi_stream_module_split_hint.kwargs:
            self._do_get_input_for_each_stream(hint_object = multi_stream_module_split_hint.kwargs,
                                        input_object = kwargs,
                                        stream_input_object = self.kwargs_streams_input[0],
                                        idx_or_key = key,
                                        stream_id = 0)
        # After we get the self.used_num_streams then we can
        # decide the inputs for the left of used_num_streams
        for stream_id in range(1, self.used_num_streams):
            # Update the split idx for current stream
            self.update_split_idx(stream_id)
            # Here we put stream go through as the outer for loop,
            # Since we assume the multi_stream_module_split_hint is not complicated to be recursive generally.
            for i in range(multi_stream_module_split_hint.args_len):
                self._do_get_input_for_each_stream(hint_object = multi_stream_module_split_hint.args,
                                                input_object = args,
                                                stream_input_object = self.args_streams_input[stream_id],
                                                idx_or_key = i,
                                                stream_id = stream_id)
            for key in multi_stream_module_split_hint.kwargs:
                self._do_get_input_for_each_stream(hint_object = multi_stream_module_split_hint.kwargs,
                                                input_object = kwargs,
                                                stream_input_object = self.kwargs_streams_input[stream_id],
                                                idx_or_key = key,
                                                stream_id = stream_id)

    def _do_generate_outputs(self, hint_object, output_object, stream_output_object, idx_or_key, stream_id):
        type_arg = type(hint_object[idx_or_key])
        if type_arg in [list]:
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_generate_outputs(hint_object[idx_or_key], output_object[idx_or_key], stream_output_object[idx_or_key], i, stream_id)
        elif type_arg in [tuple]:
            # Tuple doesn't support item change in place
            # So we change it to list for next recursion and change it back to tuple.
            temp = list(output_object[idx_or_key])
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_generate_outputs(hint_object[idx_or_key], temp, stream_output_object[idx_or_key], i, stream_id)
            output_object[idx_or_key] = tuple(temp)
        elif type_arg in [dict]:
            for key in hint_object[idx_or_key]:
                self._do_generate_outputs(hint_object[idx_or_key], output_object[idx_or_key], stream_output_object[idx_or_key], key, stream_id)
        elif (type_arg is int) or (hint_object[idx_or_key] is None):
            if hint_object[idx_or_key] is not None:
                if stream_id == 0:
                    output_object[idx_or_key] = []
                output_object[idx_or_key].append(stream_output_object[idx_or_key])
            else:
                # This object shouldn't be concat, just copy once for stream_id = 0
                if stream_id == 0:
                    output_object[idx_or_key] = stream_output_object[idx_or_key]
        else:
            assert False, "Generate outputs failed, unsupport output hint type of:{}".format(type_arg)
        return None

    def _generate_outputs(self, stream_output_object, stream_id):
        # For each position, we will push the result generated by each stream into the list
        # multi_stream_module_split_hint.args_len must be 1, since the module output will be a single output or tuple for multi outputs
        self._do_generate_outputs(hint_object = self.output_concat_hint.args,
                                output_object = self.output.args,
                                stream_output_object = stream_output_object,
                                idx_or_key = 0,
                                stream_id = stream_id)

    def _do_concat_output_for_each_stream(self, hint_object, output_object, idx_or_key):
        type_arg = type(hint_object[idx_or_key])
        if type_arg in [list]:
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_concat_output_for_each_stream(hint_object[idx_or_key], output_object[idx_or_key], i)
        if type_arg in [tuple]:
            # Tuple doesn't support item change in place
            # So we change it to list for next recursion and change it back to tuple.
            temp = list(output_object[idx_or_key])
            for i in range(hint_object[idx_or_key].__len__()):
                self._do_concat_output_for_each_stream(hint_object[idx_or_key], temp, i)
            output_object[idx_or_key] = tuple(temp)
        elif type_arg in [dict]:
            for key in hint_object[idx_or_key]:
                self._do_concat_output_for_each_stream(hint_object[idx_or_key], output_object[idx_or_key], key)
        elif (type_arg is int) or (hint_object[idx_or_key] is None):
            if hint_object[idx_or_key] is not None:
                output_object[idx_or_key] = torch.cat(output_object[idx_or_key], dim=hint_object[idx_or_key])
        else:
            assert False, "Concat output failed, unsupport output hint type of:{}".format(type_arg)
        return None

    def _concat_output_for_each_stream(self):
        # Concat the output, when here each position is already a List of tensors to be concat.
        self._do_concat_output_for_each_stream(self.output_concat_hint.args, self.output.args, 0)
        return self.output.args[0]

    def forward(self, *args, **kwargs):
        # Reset the forward status to default value which mainly contains information
        # to split inputs. They will init afterwards for each forward call.
        self.reset_forward_status()
        if self.num_streams == 1:
            # Sync execution path if num_stream is 1
            if not core.is_same_core_affinity_setting(self.core_list):
                # If the main thread's core affinity has been changed, we should set it again.
                core.pin_cpu_cores(self.core_list)
            results_raw = self.model(*args, **kwargs)
            return results_raw if self.concat_output else [results_raw]

        # Split the raw input to generate input for each stream
        self._get_input_for_each_stream(self.input_split_hint, *args, **kwargs)

        results_raw_future = []
        results_raw = []
        for stream_id in range(self.used_num_streams):
            results_raw_future.append(self.tasks[stream_id](*(self.args_streams_input[stream_id]), **(self.kwargs_streams_input[stream_id])))

        for stream_id in range(self.used_num_streams):
            # If we need to concat the output, for each position, we will push the result generated by each stream into a list for concat later.
            # For self._generate_outputs: here we put results_raw_future[stream_id].get() into a [results_raw_future[stream_id].get()]
            # to align the multi_stream_module_concat_hint structure.
            self._generate_outputs([results_raw_future[stream_id].get()], stream_id)\
                                if self.concat_output else\
                                results_raw.append(results_raw_future[stream_id].get())
        # If we need to concat the output, for each position, we will concat the result in the list (generate in self._generate_outputs).
        return self._concat_output_for_each_stream() if self.concat_output else results_raw

    def get_stream_number(self):
        return self.num_streams

class _MultiStreamBenchmarkModule(nn.Module):
    # Here is an internal Module for weight sharing benchmark
    # The diffence with MultiStreamModule:
    #    * The input will not be split. So each stream will run with the same input.
    #    * The output will not be concat. But synchronization point for each stream still exsits at the end of the forward method.
    def __init__(self, model, num_streams: Union[int, str] = "AUTO", cpu_pool: CPUPool = CPUPool(node_id=0)):
        super(_MultiStreamBenchmarkModule, self).__init__()
        assert type(cpu_pool) is CPUPool, "Input of cpu_pool must be provided with type of ipex.cpu.runtime.CPUPool"
        self.core_list = cpu_pool.core_ids
        if isinstance(num_streams, str):
            # For str input of num_streams, it must be "auto"
            if num_streams.upper() == "AUTO":
                self.num_streams = get_default_num_streams(cpu_pool) # The default selected value when auto selection is on.
            else:
                assert False, "Input of num_streams must be Number of instances or string \"AUTO\""
        else:
            assert isinstance(num_streams, int), "Input of num_streams must be Number of instances or string \"auto\""
            self.num_streams = num_streams
        if self.num_streams == 1:
            # Sync execution path if num_stream is 1.
            self.model = model
        else:
            self.cores_per_instance = self.core_list.__len__() // self.num_streams
            num_stream_allocated_extra_core = self.core_list.__len__() % self.num_streams
            self.tasks = []
            start_core_list_idx = 0
            end_core_list_idx = 0
            for j in range(self.num_streams):
                if j < num_stream_allocated_extra_core:
                    # If the core number is not divisible by stream number,
                    # the remainder streams(num_stream_allocated_extra_core) will be allocated one extra core.
                    end_core_list_idx += (self.cores_per_instance + 1)
                else:
                    end_core_list_idx += self.cores_per_instance
                self.tasks.append(Task(model, CPUPool(self.core_list[start_core_list_idx:end_core_list_idx])))
                start_core_list_idx = end_core_list_idx

    def forward(self, *args, **kwargs):
        if self.num_streams == 1:
            # Sync execution path if num_stream is 1
            if not core.is_same_core_affinity_setting(self.core_list):
                # If the main thread's core affinity has been changed, we should set it again.
                core.pin_cpu_cores(self.core_list)
            return self.model(*args, **kwargs)

        results_raw_future = []
        results_raw = []
        for j in range(self.num_streams):
            results_raw_future.append(self.tasks[j](*args, **kwargs))

        for j in range(self.num_streams):
            results_raw.append(results_raw_future[j].get())
        return results_raw[0]



from typing import List, Union

import torch
from torch import nn, Tensor
from torch.nn.modules.utils import _pair
from torch.jit.annotations import BroadcastingList2

from typing import List, Union

def _cat(tensors: List[Tensor], dim: int = 0) -> Tensor:
    """
    Efficient version of torch.cat that avoids a copy if there is only a single element in a list
    """
    # TODO add back the assert
    # assert isinstance(tensors, (list, tuple))
    if len(tensors) == 1:
        return tensors[0]
    return torch.cat(tensors, dim)


def _convert_boxes_to_roi_format(boxes: List[Tensor]) -> Tensor:
    concat_boxes = _cat([b for b in boxes], dim=0)
    temp = []
    for i, b in enumerate(boxes):
        temp.append(torch.full_like(b[:, :1], i))
    ids = _cat(temp, dim=0)
    rois = torch.cat([ids, concat_boxes], dim=1)
    return rois


def _check_roi_boxes_shape(boxes: Union[Tensor, List[Tensor]]):
    if isinstance(boxes, (list, tuple)):
        for _tensor in boxes:
            assert _tensor.size(1) == 4, \
                'The shape of the tensor in the boxes list is not correct as List[Tensor[L, 4]]'
    elif isinstance(boxes, torch.Tensor):
        assert boxes.size(1) == 5, 'The boxes tensor shape is not correct as Tensor[K, 5]'
    else:
        assert False, 'boxes is expected to be a Tensor[L, 5] or a List[Tensor[K, 4]]'
    return

def roi_align(
    input: Tensor,
    boxes: Union[Tensor, List[Tensor]],
    output_size: BroadcastingList2[int],
    spatial_scale: float = 1.0,
    sampling_ratio: int = -1,
    aligned: bool = False,
) -> Tensor:
    """
    Performs Region of Interest (RoI) Align operator with average pooling, as described in Mask R-CNN. It is optimized with parallelization and channels last support on the basis of the torchvision's roi_align.

    The semantics of Intel® Extension for PyTorch* roi_align is exactly the same as that of torchvision. We override roi_align in the torchvision with Intel® Extension for PyTorch* roi_align via ATen op registration. It is activated when Intel® Extension for PyTorch* is imported from the Python frontend or when it is linked by a C++ program. It is fully transparent to users.

    In certain cases, if you are using a self-implemented roi_align class or function that behave exactly the same as the ones in torchvision, please import the optimized one in Intel® Extension for PyTorch* as the following examples to get performance boost on Intel platforms.

    .. highlight:: python
    .. code-block:: python

        from intel_extension_for_pytorch import roi_align

    or

    .. highlight:: python
    .. code-block:: python

        from intel_extension_for_pytorch import RoIAlign

    Args:
        input (Tensor[N, C, H, W]): The input tensor, i.e. a batch with ``N`` elements. Each element
            contains ``C`` feature maps of dimensions ``H x W``.
            If the tensor is quantized, we expect a batch size of ``N == 1``.
        boxes (Tensor[K, 5] or List[Tensor[L, 4]]): the box coordinates in (x1, y1, x2, y2)
            format where the regions will be taken from.
            The coordinate must satisfy ``0 <= x1 < x2`` and ``0 <= y1 < y2``.
            If a single Tensor is passed, then the first column should
            contain the index of the corresponding element in the batch, i.e. a number in ``[0, N - 1]``.
            If a list of Tensors is passed, then each Tensor will correspond to the boxes for an element i
            in the batch.
        output_size (int or Tuple[int, int]): the size of the output (in bins or pixels) after the pooling
            is performed, as (height, width).
        spatial_scale (float): a scaling factor that maps the input coordinates to
            the box coordinates. Default: 1.0
        sampling_ratio (int): number of sampling points in the interpolation grid
            used to compute the output value of each pooled output bin. If > 0,
            then exactly ``sampling_ratio x sampling_ratio`` sampling points per bin are used. If
            <= 0, then an adaptive number of grid points are used (computed as
            ``ceil(roi_width / output_width)``, and likewise for height). Default: -1
        aligned (bool): If False, use the legacy implementation.
            If True, pixel shift the box coordinates it by -0.5 for a better alignment with the two
            neighboring pixel indices. This version is used in Detectron2

    Returns:
        Tensor[K, C, output_size[0], output_size[1]]: The pooled RoIs.
    """
    _check_roi_boxes_shape(boxes)
    rois = boxes
    output_size = _pair(output_size)
    if not isinstance(rois, torch.Tensor):
        rois = _convert_boxes_to_roi_format(rois)
    return torch.ops.torch_ipex.ROIAlign_forward(input, rois, spatial_scale,
                                           output_size[0], output_size[1],
                                           sampling_ratio, aligned)




import os
import copy
from typing import List, Dict, Tuple, Any, Optional
import torch
import torch.nn.functional as F
from torch.fx.node import map_aggregate
from torch.ao.quantization import PlaceholderObserver
from torch.quantization.qconfig import QConfig

from ._utils import get_torch_function_hook_type, HookType, get_module_hook_type, OpQuantizeabilityType, \
    attach_op_convert_info_to_model, save_quant_state, attach_scale_zp_values_to_model, convert_quant_state_map_to_nodes, \
        sync_pool_and_lstm_input_output_scale_zp, module_call_to_function_call, quantized_modules_has_weights, \
        load_qconf_summary_to_model, get_fqn_valid_for_module_dict_key
from ._quantization_state import AutoQuantizationState, AutoQuantizationStateModuleDict, init_model_quant_state
from ._recipe import get_default_recipe
from ._module_swap_utils import swap_child_modules

# AutoQuantizationState lives in parent module's _modules.
# Currently, `torch.nn.Sequential`'s forward iterates over all
# items in _modules. To avoid changing the meaning of the program, for
# now we patch the forward to ignore our quantization state.
# Note: this is a hackedy hack, before launching we should consider
# checking the fix into `torch.nn.Sequential` to avoid the patch.
def _nn_sequential_patched_forward(cls, input):
    for module in cls:
        if not isinstance(module, AutoQuantizationStateModuleDict):
            input = module(input)
    return input

def _check_add_has_scalar_input(args):
    r"""
    This function is about check add whether has scalar input.
    """
    for arg in args:
        if not isinstance(arg, torch.Tensor):
            return True
    return False

def auto_prepare(
    model : torch.nn.Module,
    configure: QConfig,
    example_inputs: Tuple[Any],
) -> torch.nn.Module:

    def convert_to_interception_proxy(x):
        if isinstance(x, torch.Tensor):
            return x.as_subclass(QuantizationPrepareTensorProxy)  # type: ignore[arg-type]
        else:
            return x

    cur_module = None
    first_call = True
    module_stack : List[torch.nn.Module] = []
    # Counter for tensor IDs, will be modified inplace by quant state.
    # This is used to track tensors from output ops to input ops. For example,
    # if op_n had a tensor output with id=1, and op_n+2 had a tensor input with
    # id=1, we know that the output of op_n is the input to op_n+2. Note,
    # this is a list because it needs to incremented inplace.
    qtensor_id = [0]
    module_id_to_fqn: Dict[int, str] = {}

    # Counter for global quantizeable ops, useful for intermediate activation
    # logging.
    global_op_idx = [0]

    global_disable_torch_function_override = False

    class QuantizationPrepareTensorProxy(torch.Tensor):
        """
        An override of `torch.Tensor` to enable dynamic tracing for
        quantization.
        For each function with a `__torch_function__` override, this proxy does
        the following for functions which need quantization:
        1. calls `_auto_quant_state.validate_cur_op` to validate that
           the currently seen op is the same as what was recorded during tracing
        2. calls `_auto_quant_state.op_prepare_before_hook`
        3. executes the original function
        4. calls `_auto_quant_state.op_prepare_after_hook`
        5. calls `_auto_quant_state.mark_cur_op_complete` to increment
           the current op index in preparation for the next op
        Otherwise, calls the original function.
        """

        @classmethod
        def __torch_function__(cls, func, types, args=(), kwargs=None):
            nonlocal global_disable_torch_function_override
            if (
                # global override means disable the override here
                global_disable_torch_function_override or
                # to prevent printing things from going into an infinite loop
                func == torch.Tensor.__repr__ or
                # we don't need to override getters in this framework
                func.__name__ == '__get__'
            ):
                return super().__torch_function__(func, types, args, kwargs)

            # if we are in a function, the current module is always a parent
            nonlocal cur_module
            parent_module = cur_module
            nonlocal qtensor_id
            kwargs = kwargs if kwargs else {}
            hook_type = get_torch_function_hook_type(parent_module, func)
            # Don't support torch.add(tensor, scaler)
            # case, scalar+scalar, pytorch trace will convert the first input as a tensor at convert step,
            # but we didn't collect the quant info at calibration step, which can't get
            # quant info here(asster KeyError), so we disable torch.add(tensor, scaler) quantizaiton.
            if hook_type is HookType.OP_HOOKS and func in [torch.add, torch.Tensor.add] and _check_add_has_scalar_input(args):
                hook_type = None

            if hook_type is HookType.OP_HOOKS:
                fqn = module_id_to_fqn[id(parent_module)] if parent_module else None
                qstate = parent_module._auto_quant_state  # type: ignore[attr-defined]
                if not first_call:
                    qstate.validate_cur_op(func)
                # run "before" hook
                if first_call:
                    args, kwargs = qstate.first_call_op_prepare_before_hook(
                        func, args, kwargs, qtensor_id, fqn, parent_module, OpQuantizeabilityType.QUANTIZEABLE)
                else:
                    args, kwargs = qstate.op_prepare_before_hook(
                        func, args, kwargs)
                # forward
                output = super().__torch_function__(func, types, args, kwargs)
                # run "after" hook
                if first_call:
                    output = qstate.first_call_op_prepare_after_hook(
                        func, output, args, qtensor_id, OpQuantizeabilityType.QUANTIZEABLE)
                else:
                    output = qstate.op_prepare_after_hook(
                        func, output, args, global_op_idx)
                qstate.mark_cur_op_complete(func)
            else:
                # Hook type is not HookType.OP_HOOKS, if first_call is True we
                # record the DAG of non-quantizeable ops.
                if first_call:
                    qstate = getattr(parent_module, '_auto_quant_state', None)
                    if qstate:
                        fqn = module_id_to_fqn.get(id(parent_module), None) \
                            if parent_module else None
                        args, kwargs = qstate.first_call_op_prepare_before_hook(
                            func, args, kwargs, qtensor_id, fqn, parent_module, OpQuantizeabilityType.NOT_QUANTIZEABLE)

                output = super().__torch_function__(func, types, args, kwargs)

                if first_call:
                    qstate = getattr(parent_module, '_auto_quant_state', None)
                    if qstate:
                        output = qstate.first_call_op_prepare_after_hook(
                            func, output, args, qtensor_id, OpQuantizeabilityType.NOT_QUANTIZEABLE)

            if output is NotImplemented:
                with torch._C.DisableTorchFunction():
                    output = func(*args, **kwargs).as_subclass(
                        QuantizationPrepareTensorProxy)
                assert output is not NotImplemented

            return output

        def __repr__(self):
            return f'QuantizationPrepareTensorProxy({super().__repr__()})'

        # TODO(future PR): add other math overrides

    class QuantizationInterceptionModule(type(model)):  # type: ignore[misc]
        """
        An override of user defined subclass of `nn.Module` to enable
        dynamic tracing for quantization.
        `cur_module` keeps track of the current module in the stack.
        During the fist call, an `AutoQuantizationState` object is created and
        attached to each non-leaf modules which we need to check for
        quantizeable operations.
        We override the `__call__` function to do the following for each
        module:
        If the module is an op which needs quantization:
        1. calls `_auto_quant_state.validate_cur_op` to validate that
           the currently seen op is the same as what was recorded during tracing
        2. calls parent module's `._auto_quant_state.op_prepare_before_hook`
        3. executes the original module forward
        4. calls parent module's `_auto_quant_state.op_prepare_after_hook`
        5. calls `_auto_quant_state.mark_cur_op_complete` to increment
           the current op index in preparation for the next op
        Otherwise, calls the original module forward.
        """

        def __call__(self, *args, **kwargs):
            new_args = map_aggregate(args, convert_to_interception_proxy)
            new_kwargs = map_aggregate(kwargs, convert_to_interception_proxy)
            orig_module_call = torch.nn.Module.__call__
            orig_nn_sequential_forward = torch.nn.Sequential.forward

            def _patched_module_call(self, *args, **kwargs):
                nonlocal cur_module
                old_module = cur_module
                cur_module = self
                try:
                    parent_module = module_stack[-1] if len(module_stack) else None
                    module_stack.append(self)
                    fqn = module_id_to_fqn.get(id(self), None)

                    hook_type = get_module_hook_type(parent_module, cur_module)
                    if hook_type is HookType.OP_HOOKS:
                        parent_qstate: AutoQuantizationState = \
                            parent_module._auto_quant_state  # type: ignore[union-attr, assignment]
                        # before hooks
                        if not first_call:
                            parent_qstate.validate_cur_op(cur_module)

                        # If we are in this hook, `cur_module` is a leaf module.
                        # Therefore, we do not need to override any of its
                        # children. Disabling the overrides for performance.
                        nonlocal global_disable_torch_function_override
                        old_global_disable_torch_function_override = \
                            global_disable_torch_function_override
                        global_disable_torch_function_override = True

                        if first_call:
                            # mypy ignore is used instead of assert because this
                            # runs on every forward and assert has a performance cost
                            args, kwargs = parent_qstate.first_call_op_prepare_before_hook(
                                cur_module, args, kwargs, qtensor_id,
                                fqn, cur_module, # type: ignore[arg-type]
                                OpQuantizeabilityType.QUANTIZEABLE)
                        else:
                            # mypy ignore is used instead of assert because this
                            # runs on every forward and assert has a performance cost
                            args, kwargs = parent_qstate.op_prepare_before_hook(
                                cur_module, args, kwargs)  # type: ignore[arg-type]

                        # original forward
                        output = orig_module_call(self, *args, **kwargs)
                        # Re-enable the overrides.
                        global_disable_torch_function_override = \
                            old_global_disable_torch_function_override

                        # after hooks
                        if first_call:
                            output = parent_qstate.first_call_op_prepare_after_hook(
                                cur_module, output, args, qtensor_id, OpQuantizeabilityType.QUANTIZEABLE)
                        else:
                            output = parent_qstate.op_prepare_after_hook(
                                cur_module, output, args, global_op_idx)
                        parent_qstate.mark_cur_op_complete(cur_module)
                    elif hook_type is HookType.MODULE_IO_HOOKS:
                        cur_qstate = cur_module._auto_quant_state
                        cur_qstate.reset_to_new_call()
                        # original forward
                        output = orig_module_call(self, *args, **kwargs)

                        # after hooks
                        if first_call:
                            output = cur_qstate.first_call_outputs_prepare_hook(
                                output, qtensor_id)
                        else:
                            output = cur_qstate.outputs_prepare_hook(output)

                        cur_qstate.validate_is_at_last_seen_idx()
                    elif hook_type is HookType.ARG_DEQUANTS:
                        if first_call and parent_module is not None:
                            parent_qstate_fc = getattr(
                                parent_module, '_auto_quant_state', None)
                            if parent_qstate_fc:
                                args, kwargs = \
                                    parent_qstate_fc.first_call_op_prepare_before_hook(
                                        cur_module, args, kwargs, qtensor_id, fqn,
                                        cur_module,
                                        OpQuantizeabilityType.NOT_QUANTIZEABLE)

                        output = orig_module_call(self, *args, **kwargs)
                        # if this fp32 was inplace, make sure to set the output dtype
                        # back to torch.float
                        if hasattr(output, '_qtensor_info'):
                            del output._qtensor_info

                        if first_call and parent_module is not None:
                            parent_qstate_fc = getattr(
                                parent_module, '_auto_quant_state', None)
                            if parent_qstate_fc:
                                output = \
                                    parent_qstate_fc.first_call_op_prepare_after_hook(
                                        cur_module, output, args, qtensor_id,
                                        OpQuantizeabilityType.NOT_QUANTIZEABLE)
                    else:
                        output = orig_module_call(self, *args, **kwargs)

                    return output
                finally:
                    module_stack.pop()
                    cur_module = old_module

            torch.nn.Module.__call__ = _patched_module_call
            torch.nn.Sequential.forward = _nn_sequential_patched_forward  # type: ignore[assignment]
            nonlocal first_call
            try:
                if first_call:
                    init_model_quant_state(self, module_id_to_fqn, configure)
               
                global_op_idx[0] = 0
                output = super().__call__(*new_args, **new_kwargs)

                if first_call:
                    for _, v in self.named_modules():
                        if hasattr(v, '_auto_quant_state'):
                            v._auto_quant_state.insert_observers(v)
                return output
            finally:
                torch.nn.Module.__call__ = orig_module_call
                torch.nn.Sequential.forward = orig_nn_sequential_forward  # type: ignore[assignment]
                first_call = False

        def save_qconf_summary(self, qconf_summary):
            r"""
            This function is about save model's quant_state_map to a json file.
            """
            assert qconf_summary is not None, "A configure file name should be given to save the qconf_summary"
            quant_state_map = self._fqn_to_auto_quant_state_map
            # If user have given a json file, we will save the qconf_summary according to the user's setting,
            # otherwise,  we will first get a default_recipe, and then save the default_recipe's setting.
            if not hasattr(self, '_qconf_summary'):
                # compute scales and zero_point.
                attach_scale_zp_values_to_model(model)
                nodes = convert_quant_state_map_to_nodes(quant_state_map)
                # pooling and lstm's input and output should have same scale_zp.
                sync_pool_and_lstm_input_output_scale_zp(quant_state_map, nodes)
                get_default_recipe(nodes)
            # Setting model qconf_summary attr which can be easily to check the whether the scale/zp has been computed.
            self._qconf_summary = qconf_summary
            save_quant_state(quant_state_map, qconf_summary)
        
        def load_qconf_summary(self, qconf_summary):
            r"""
            This function is about load the user qconf_summary, which will overwrite the model's quant_state_map.
            """
            if (os.path.exists(qconf_summary) and os.stat(qconf_summary).st_size != 0):
                self._qconf_summary = qconf_summary
                load_qconf_summary_to_model(self, qconf_summary)
            else:
                assert False, "Can not load a empty file or none existed file" + qconf_summary

    model.q_config = configure
    # For Dynamic quantization, most user model has a dynamic control flow, the DBR
    # doesn't support it now, so there skip DRB when user want to run dynamic quantization. 
    if not isinstance(configure.activation(), PlaceholderObserver):
        model.__class__ = QuantizationInterceptionModule
        # init model quantization state using example_inputs
        model(*example_inputs)
    return model

def copy_prepared_model(model):
    copied_model = copy.deepcopy(model)
    copied_model.q_config = model.q_config
    if isinstance(copied_model.q_config.activation(), PlaceholderObserver):
        return copied_model
    copied_model._fqn_to_auto_quant_state_map = copy.deepcopy(model._fqn_to_auto_quant_state_map)
    named_modules = list(copied_model.named_modules())
    for fqn, v in named_modules:
        fqn_to_use_for_key = get_fqn_valid_for_module_dict_key(fqn)
        if fqn_to_use_for_key in copied_model._fqn_to_auto_quant_state_map:
            auto_quant_state = copied_model._fqn_to_auto_quant_state_map[fqn_to_use_for_key]
            object.__setattr__(v, '_auto_quant_state', auto_quant_state)
    if hasattr(model, '_qconf_summary'):
        copied_model._qconf_summary = copy.deepcopy(model._qconf_summary)
    copied_model.__class__ = model.__class__
    return copied_model

def auto_convert(
    module : torch.nn.Module,
    ) -> torch.nn.Module:
    def convert_to_dispatch_proxy(x):
        if isinstance(x, torch.Tensor):
            return x.as_subclass(QuantizationConvertTensorProxy)  # type: ignore[arg-type]
        else:
            return x

    global_disable_torch_function_override = False

    class QuantizationConvertTensorProxy(torch.Tensor):
        """
        An override of `torch.Tensor` to enable dynamic dispatch for
        quantization inference.
        For each function with a `__torch_fuction__` override, this proxy does
        the following for functions which need quantization:
        1. calls `_auto_quant_state.validate_cur_op` to validate that
           the currently seen op is the same as what was recorded during tracing
        2. calls `_auto_quant_state.op_convert_before_hook`.
        3. executes the function, with target, args and kwargs possibly modified
           by (2)
        4. calls `_auto_quant_state.inference_function_after_hook`.
        5. calls `_auto_quant_state.mark_cur_op_complete` to increment
           the current op index in preparation for the next op
        Otherwise, calls the original function.
        """

        @classmethod
        def __torch_function__(cls, func, types, args=(), kwargs=None):
            nonlocal global_disable_torch_function_override
            if (
                # global override means disable the override here
                global_disable_torch_function_override or
                # to prevent printing things from going into an infinite loop
                func == torch.Tensor.__repr__ or
                # we don't need to override getters in this framework
                func.__name__ == '__get__'
            ):
                return super().__torch_function__(func, types, args, kwargs)

            kwargs = kwargs if kwargs else {}
            # if we are in a function, the current module is always a parent
            parent_module = cur_module
            hook_type = get_torch_function_hook_type(parent_module, func)
            # Don't support torch.add(tensor, scaler)
            # case, scalar+scalar, pytorch trace will convert the first input as a tensor,
            # but we didn't collect the quant info at calibration step, which can't get
            # quant info here(asster KeyError), so we disable torch.add(tensor, scaler) quantizaiton.
            if hook_type is HookType.OP_HOOKS and func in [torch.add, torch.Tensor.add] and _check_add_has_scalar_input(args):
                hook_type = None

            if hook_type is HookType.OP_HOOKS:
                qstate: AutoQuantizationState = parent_module._auto_quant_state  # type: ignore[union-attr]
                # before hooks
                qstate.validate_cur_op(func)
                func, args, kwargs = qstate.op_convert_before_hook(
                    func, args, kwargs, parent_module)  # type: ignore[arg-type]

                # forward
                output = super().__torch_function__(func, types, args, kwargs)

                # after hooks
                output = qstate.op_convert_after_hook(
                    func, output)
                qstate.mark_cur_op_complete(func)
            else:  # HookType.NONE
                output = super().__torch_function__(func, types, args, kwargs)

            if output is NotImplemented:
                with torch._C.DisableTorchFunction():
                    output = func(*args, **kwargs).as_subclass(
                        QuantizationConvertTensorProxy)
                assert output is not NotImplemented
            return output

        def __repr__(self):
            return f'QuantizationConvertTensorProxy({super().__repr__()})'

    cur_module = None
    module_stack : List[torch.nn.Module] = []

    assert len(module.__class__.__bases__) == 1

    class QuantizationDispatchModule(module.__class__.__bases__[0]):  # type: ignore[name-defined]
        """
        An override of user defined subclass of `nn.Module` to enable
        dynamic tracing for quantization, after model conversion
        to quantized domain.
        `cur_module` keeps track of the current module in the stack.
        Tensor arguments are converted to `QuantizationConvertTensorProxy`.
        We override the `__call__` function to do the following for each
        module:
        If the module is an op which needs quantization:
        1. calls `_auto_quant_state.validate_cur_op` to validate that
           the currently seen op is the same as what was recorded during tracing
        2. calls parent module's `._auto_quant_state.op_convert_before_hook`
        3. executes the original module forward
        4. calls parent module's `_auto_quant_state.op_convert_after_hook`
        5. calls `_auto_quant_state.mark_cur_op_complete` to increment
           the current op index in preparation for the next op
        Otherwise, calls the original module forward.
        """

        def __call__(self, *args, **kwargs):
            new_args = map_aggregate(args, convert_to_dispatch_proxy)
            new_kwargs = map_aggregate(kwargs, convert_to_dispatch_proxy)
            orig_module_call = torch.nn.Module.__call__
            orig_nn_sequential_forward = torch.nn.Sequential.forward

            def _patched_module_call(self, *args, **kwargs):
                nonlocal cur_module
                old_module = cur_module
                cur_module = self
                nonlocal global_disable_torch_function_override
                try:
                    parent_module = module_stack[-1] if len(module_stack) else None
                    module_stack.append(self)
                    hook_type = get_module_hook_type(parent_module, cur_module)
                    if hook_type is HookType.OP_HOOKS:
                        # before hooks
                        qstate: AutoQuantizationState = \
                            parent_module._auto_quant_state  # type: ignore[union-attr, assignment]
                        qstate.validate_cur_op(cur_module)

                        # If we are in this hook, `cur_module` is a leaf module.
                        # Therefore, we do not need to override any of its
                        # children. Disabling the overrides for performance.
                        old_global_disable_torch_function_override = \
                            global_disable_torch_function_override
                        global_disable_torch_function_override = True
                        _, args, kwargs = qstate.op_convert_before_hook(
                            cur_module, args, kwargs, cur_module)
                        if type(cur_module) in quantized_modules_has_weights:
                            weights = qstate.op_weight_convert_before_hook(cur_module)
                            output = module_call_to_function_call(self, args, weights)
                        else:
                             output = orig_module_call(self, *args, **kwargs)
                        # after hooks
                        output = qstate.op_convert_after_hook(
                            cur_module, output)

                        # Re-enable the override.
                        global_disable_torch_function_override = \
                            old_global_disable_torch_function_override

                        qstate.mark_cur_op_complete(cur_module)
                    elif hook_type is HookType.MODULE_IO_HOOKS:
                        cur_qstate: AutoQuantizationState = cur_module._auto_quant_state
                        cur_qstate.reset_to_new_call()
                        # before hooks (TODO)
                        # forward
                        output = orig_module_call(self, *args, **kwargs)
                        # after hooks
                        # For the sake of performance, we assume no overrides
                        # are needed for quantizing/dequantizing things
                        old_global_disable_torch_function_override = \
                            global_disable_torch_function_override
                        global_disable_torch_function_override = True
                        
                        output = cur_qstate.outputs_convert_hook(output)
                        global_disable_torch_function_override = \
                            old_global_disable_torch_function_override
                        cur_qstate.validate_is_at_last_seen_idx()
                    else:
                        output = orig_module_call(self, *args, **kwargs)
                    return output
                finally:
                    module_stack.pop()
                    cur_module = old_module

            torch.nn.Module.__call__ = _patched_module_call
            torch.nn.Sequential.forward = _nn_sequential_patched_forward  # type: ignore[assignment]

            try:
                output = super().__call__(*new_args, **new_kwargs)

                def unwrap_proxy(a):
                    if isinstance(a, QuantizationConvertTensorProxy):
                        a.__class__ = torch.Tensor  # type: ignore[assignment]
                    return a

                output = map_aggregate(output, unwrap_proxy)
                return output
            finally:
                torch.nn.Module.__call__ = orig_module_call
                torch.nn.Sequential.forward = orig_nn_sequential_forward  # type: ignore[assignment]
 
    # If module doesn't have a configure_file attr, we can say that user has run save_qconf_summary method which have
    # computed the scales and zp, or use the user's setting from a given json file(load_qconf_summary), we need to compute
    # the scale and zp here.
    if not hasattr(module, '_qconf_summary'):
        quant_state_map = module._fqn_to_auto_quant_state_map
        # compute scales and zero_point.
        attach_scale_zp_values_to_model(module)
        nodes = convert_quant_state_map_to_nodes(quant_state_map)
        # pooling and lstm's input and output should have same scale_zp.
        sync_pool_and_lstm_input_output_scale_zp(quant_state_map, nodes)
        get_default_recipe(nodes)
    else:
        # Clear observer if module have, this will works when the user's json setting is loaded.
        for _, v in module._fqn_to_auto_quant_state_map.items():
            v.tensor_id_to_observer.clear()
            v.weight_tensor_id_to_observer.clear()
    # Attach quant_info to parent each module
    attach_op_convert_info_to_model(module)
    swap_child_modules(module)
    module.__class__ = QuantizationDispatchModule
    return module



import torch
import intel_extension_for_pytorch as ipex  # noqa F401
from ..utils._weight_prepack import _IPEXLinear as _IPEXLinear
import enum

class EltwiseType(enum.IntEnum):
    NotFused = 0
    ReLU = 1
    Sigmoid = 2

class IPEXLinearEltwise(torch.nn.Module):

    def __init__(self, ipex_linear_module, eltwise='relu'):
        super(IPEXLinearEltwise, self).__init__()
        assert isinstance(ipex_linear_module, _IPEXLinear)
        self.m = ipex_linear_module
        if eltwise == 'relu':
            self.eltwise = EltwiseType.ReLU
        else:
            assert eltwise == 'sigmoid'
            self.eltwise = EltwiseType.Sigmoid

    def forward(self, x):
        return torch.ops.torch_ipex.ipex_linear_eltwise(
            x, self.m.weight, self.m.bias, self.eltwise, self.m.ctx.get_data_handle())



import torch
import torch.fx as fx
import types

def override_is_leaf_module():
    fx_tracer = fx.Tracer
    orig_is_leaf_module_fn = fx_tracer.is_leaf_module
    def ipex_is_leaf_module_fn(self, m: torch.nn.Module, module_qualified_name : str) -> bool:
        is_ipex = m.__module__.startswith('intel_extension_for_pytorch.nn')
        return is_ipex or orig_is_leaf_module_fn(self, m, module_qualified_name)
    setattr(fx_tracer, 'is_leaf_module', types.MethodType(ipex_is_leaf_module_fn, fx_tracer))

override_is_leaf_module()



import torch
import copy
import warnings

from torch.nn.utils.rnn import PackedSequence

class _LSTM(torch.nn.LSTM):
    # This is a solution to swap the lstm module with the ipex counterpart
    # and will upstream this operator to PyTorch when oneDNN support
    # bias and src_iter_c in bf16 in bf16 inference. Will keep this
    # for better support of blocked-format weight, e.g. for training.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # port from torch/nn/modules/rnn.py
    # replace the _VF.lstm with torch.ops.torch_ipex.lstm when the input is not PackedSequence
    def forward(self, input, hx=None):  # noqa: F811
        orig_input = input
        # xxx: isinstance check needs to be in conditional for TorchScript to compile
        if isinstance(orig_input, PackedSequence):
            # fallback to PyTorch LSTM since PackedSequence unsupported in oneDNN
            return super(LSTM, self).forward(input, hx)
        else:
            batch_sizes = None
            max_batch_size = input.size(0) if self.batch_first else input.size(1)
            sorted_indices = None
            unsorted_indices = None

        if hx is None:
            num_directions = 2 if self.bidirectional else 1
            real_hidden_size = self.proj_size if self.proj_size > 0 else self.hidden_size
            h_zeros = torch.zeros(self.num_layers * num_directions,
                                  max_batch_size, real_hidden_size,
                                  dtype=input.dtype, device=input.device)
            c_zeros = torch.zeros(self.num_layers * num_directions,
                                  max_batch_size, self.hidden_size,
                                  dtype=input.dtype, device=input.device)
            hx = (h_zeros, c_zeros)
        else:
            # Each batch of the hidden state should match the input sequence that
            # the user believes he/she is passing in.
            hx = self.permute_hidden(hx, sorted_indices)

        self.check_forward_args(input, hx, batch_sizes)
        result = torch.ops.torch_ipex.ipex_lstm(input, hx, self._flat_weights, self.bias, self.num_layers,
                        self.dropout, self.training, self.bidirectional, self.batch_first)
        output = result[0]
        hidden = result[1:]

        return output, self.permute_hidden(hidden, unsorted_indices)

def replace_lstm_with_ipex_lstm(model):
    # replace lstm with ipex lstm during inference
    # does not support the case where model itself is torch.nn.LSTM
    for child_name, child in model.named_children():
        if isinstance(child, torch.nn.LSTM):
            assert hasattr(child, "weight_ih_l0"), "torch.nn.LSTM should have weight_ih_l0"
            ipex_lstm = _LSTM(child.input_size, child.hidden_size,
                child.num_layers, child.bias, child.batch_first,
                child.dropout, child.bidirectional, child.proj_size,
                child.weight_ih_l0.device, child.weight_ih_l0.dtype)
            ipex_lstm.__dict__ = copy.deepcopy(child.__dict__)
            setattr(model, child_name, ipex_lstm)
        else:
            replace_lstm_with_ipex_lstm(child)

def replace_dropout_with_identity(model):
    # replace dropout with identity during inference, so that aten::dropout won't be on the JIT graph.
    # This optimization may provide more fusion opportunites on the graph.
    if isinstance(model, torch.jit.ScriptModule):
        return
    if not model.training:
        for child_name, child in model.named_children():
            if isinstance(child, torch.nn.Dropout):
                setattr(model, child_name, torch.nn.Identity())
            else:
                replace_dropout_with_identity(child)

def convert_module_data_type(module, dtype):
    # convert weights(bias) of module to dtype to reduce dtype reorder
    module_convert_list = [torch.nn.Conv2d,
                           torch.nn.Conv3d,
                           torch.nn.ConvTranspose2d,
                           torch.nn.ConvTranspose3d,
                           torch.nn.Linear,
                           torch.nn.Embedding,
                           torch.nn.LSTM]
    for module_cls in module_convert_list:
        if isinstance(module, module_cls):
            if module_cls is torch.nn.LSTM:
                for name, param in module.named_parameters():
                    ori_data = getattr(getattr(module, name), "data")
                    ori_data_dtype = ori_data.dtype
                    if ori_data_dtype == torch.float or ori_data_dtype == torch.bfloat16:
                        casted_data = ori_data.detach().clone().to(dtype)
                        setattr(getattr(module, name), "data", casted_data)
                    else:
                        warnings.warn(f"WARNING: Can't convert model's parameters dtyep from {ori_data_dtype} to {dtype}")
                        break
            else:
                ori_data_dtype = module.weight.dtype
                # Assume weight and bias have same dtype, only need check weight dtype here.
                if ori_data_dtype == torch.float or ori_data_dtype == torch.bfloat16:
                    weight_data = module.weight.detach().clone().to(dtype)
                    module.weight.data = weight_data
                    if hasattr(module, 'bias') and module.bias is not None:
                        bias_data = module.bias.detach().clone().to(dtype)
                        module.bias.data = bias_data
                else:
                    warnings.warn(f"WARNING: Can't convert model's parameters dtype from {ori_data_dtype} to {dtype}")
            break
    for child in module.children():
        convert_module_data_type(child, dtype)
    return module




from typing import Dict, Callable, Any, Optional

import torch
import torch.nn as nn

from torch.ao.quantization import swap_module
import torch.nn.quantized.dynamic as nnqd
from torch.quantization.qconfig import QConfig

# Default map for swapping dynamic modules
DEFAULT_DYNAMIC_QUANT_MODULE_MAPPINGS : Dict[Callable, Any] = {
    nn.Linear: nnqd.Linear,
    nn.LSTM: nnqd.LSTM,
    # TODO: support more RNN module
    #nn.GRUCell: nnqd.GRUCell,
    #nn.GRU: nnqd.GRU,
    #nn.LSTMCell: nnqd.LSTMCell,
    #nn.RNNCell: nnqd.RNNCell,
}

def _get_qconfig_dtypes(qconfig):
    r"""
    Returns the qconfig tuple for qconfig:
    (activation_dtype, weight_dtype, activation_compute_dtype)
    """
    assert qconfig is not None
    activation = qconfig.activation()
    weight = qconfig.weight()
    compute_dtype = activation.compute_dtype if hasattr(activation, 'compute_dtype') else None
    return (activation.dtype, weight.dtype, compute_dtype)

def _op_is_int8_dynamically_quantized(qconfig) -> bool:
    r""" 
    Given a qconfig, returns True if this op is using int8 dynamic
    quantization
    """
    activation_dtype, weight_dtype, activation_compute_dtype = \
        _get_qconfig_dtypes(qconfig)
    return (
        activation_dtype is torch.float and
        # for now, the lines below assume fbgemm or qnnpack
        weight_dtype is torch.qint8 and
        activation_compute_dtype is torch.quint8
    )

def _swap_child_modules(
    module: torch.nn.Module,
    fqn_qconfig: Dict[str, QConfig],
    dynamic_mappings: Dict[Callable, Any] = DEFAULT_DYNAMIC_QUANT_MODULE_MAPPINGS,
    parent_fqn: Optional[str] = None,
) -> None:
    """
    For each direct child of `module`, swaps it using `dyanamic_mappings`
    if the qconfig for that child is using int8 dynamic quantization,
    and the module type is in the mapping.
    Recursively calls itself on each child.
    """
    reassign = {}
    for local_fqn, mod in module.named_children():
        if parent_fqn is None:
            global_fqn = local_fqn
        else:
            global_fqn = f"{parent_fqn}.{local_fqn}"
        _swap_child_modules(mod, fqn_qconfig, dynamic_mappings, global_fqn)

        if global_fqn in fqn_qconfig:
            qconfig = fqn_qconfig[global_fqn]
            if not qconfig:
                continue
            mod.qconfig = qconfig
            op_int8_dynamically_quantized = _op_is_int8_dynamically_quantized(qconfig)
            if op_int8_dynamically_quantized:
                if not type(mod) in dynamic_mappings:
                    continue 
                reassign[local_fqn] = swap_module(mod, dynamic_mappings, {})

    for key, value in reassign.items():
        module._modules[key] = value


def swap_child_modules(
    module: torch.nn.Module,
    dynamic_mappings: Dict[Callable, Any] = DEFAULT_DYNAMIC_QUANT_MODULE_MAPPINGS,
) -> None:
    fqn_qconfig = {}
    for _, v in module._fqn_to_auto_quant_state_map.items():
        if len(v.idx_to_seen_q_op_infos) > 0:
            for _, op_info in v.idx_to_seen_q_op_infos.items():
                fqn_qconfig[op_info.fqn] = op_info.qconfig

    _swap_child_modules(module, fqn_qconfig, dynamic_mappings)



import torch
from torch.ao.quantization import PlaceholderObserver, PerChannelMinMaxObserver, HistogramObserver, QConfig


_default_weight_observer = PerChannelMinMaxObserver.with_args(dtype=torch.qint8, qscheme=torch.per_channel_symmetric)

default_static_qconfig = QConfig(activation=HistogramObserver.with_args(reduce_range=False),
                                 weight=_default_weight_observer)
"""
Default qconfig configuration for static quantization.
"""

default_dynamic_qconfig = QConfig(activation=PlaceholderObserver.with_args(dtype=torch.float, compute_dtype=torch.quint8),
                                  weight=_default_weight_observer)
"""
Default qconfig configuration for dynamic quantization.
"""



import enum
import json
from collections import OrderedDict
from typing import Callable, Optional
import inspect
import numbers

import torch
import torch.nn as nn
from torch import _VF
import torch.nn.functional as F
from torch.quantization.qconfig import QConfig
from intel_extension_for_pytorch.nn.functional import interaction

from ._quantization_state_utils import QTensorInfo


add_and_mul_ops = set([
    torch.add,
    torch.Tensor.add,
    # TODO, mul op?
])

quantized_modules_has_weights = set([
    torch.nn.Conv2d,
    torch.nn.Conv3d,
    torch.nn.Linear,
    torch.nn.EmbeddingBag,
    torch.nn.ConvTranspose2d,
    torch.nn.ConvTranspose3d,
    torch.nn.LSTM,
    ])

# those ops only support int8->int8, not int8->fp32/bf16
int8_int8_ops = set([
    str(F.adaptive_avg_pool2d),
    str(F.adaptive_avg_pool3d),
    str(F.avg_pool2d),
    str(F.avg_pool3d),
    str(F.max_pool2d),
    str(F.max_pool3d),
    str(nn.MaxPool2d),
    str(nn.MaxPool3d),
    str(nn.AvgPool2d),
    str(nn.AvgPool3d),
    str(nn.AdaptiveAvgPool2d),
    str(nn.AdaptiveAvgPool3d),
    str(torch.Tensor.relu),
    str(torch.relu),
    str(F.relu),
    str(nn.ReLU),
    str(torch.flatten),
    str(torch.Tensor.flatten),
    str(torch.nn.Flatten),
    # the following op will be supported at next step.
    #str(torch.Tensor.sigmoid),
    #str(torch.sigmoid),
    #str(F.sigmoid),
    #str(nn.Sigmoid),
    #str(F.gelu),
    #str(nn.GELU),
    # ipex customer op
    str(interaction),
    str(torch.ops.torch_ipex.interaction_forward),
    str(torch.embedding_bag),
    str(F.embedding_bag),
    str(torch.nn.EmbeddingBag),
    str(torch.nn.LSTM),
])

class OpQuantizeabilityType(enum.Enum):
    QUANTIZEABLE = 0
    NOT_QUANTIZEABLE = 1

class FuncOutputObsType(enum.Enum):
    NONE = 0
    NEW_OBS = 1
    REUSES_FIRST_INPUT_OBS = 2

# quantizeable modules
def is_leaf(
    m: torch.nn.Module,
) -> bool:

    # TODO(future PR): extend to the rest of the container classes
    container_classes = (
        torch.nn.Sequential,
        torch.nn.ModuleList,
    )
    return (
        # allowlist everything in torch.nn except containers
        (m.__module__.startswith('torch.nn') and (
            not isinstance(m, container_classes)
        ))
    )

def get_fqn_valid_for_module_dict_key(fqn: str) -> str:
    """
    Modifies `fqn`(fully qualified name) to make it a valid key to a ModuleDict.
    """
    if fqn == '':
        fqn = ' '
    return fqn.replace('.', ':')

class HookType(enum.Enum):
    """
    Describes the various types of function and module hooks that are used
    to implement quantization syntax transforms.
    """
    # Hooks which are run before, during and after a quantizeable op.
    # Usually used for op input and output observation, subsituating
    # quantized kernels, and dynamically looking up arguments to quantized
    # kernels.
    OP_HOOKS = 0
    # Hooks which are run before or after a `torch.nn.Module` which
    # is a non-leaf. Usually used for dtype transforms if the user requests
    # that the inputs or outputs of a certain module are of some dtype.
    MODULE_IO_HOOKS = 1
    # Hooks which are run before a non-quantizeable op which requires
    # `torch.float` inputs. Any inputs which are not floats are converted
    # back to floats.
    ARG_DEQUANTS = 2
    # Everything else
    NONE = 2

def get_torch_function_hook_type(
    parent_module: Optional[torch.nn.Module],
    func: Callable,
) -> HookType:
    # the direct __dict__ accesses are for performance, because
    # the default `torch.nn.Module.__getattr__` has overhead.
    parent_module_has_qstate = parent_module is not None and \
        '_auto_quant_state' in parent_module.__dict__
    needs_op_hooks = parent_module_has_qstate and \
        parent_module.__dict__['_auto_quant_state'].cur_op_needs_hooks(func)  # type: ignore[union-attr, operator]

    if needs_op_hooks:
        return HookType.OP_HOOKS
    elif (
        parent_module_has_qstate
    ):
        return HookType.ARG_DEQUANTS
    else:
        return HookType.NONE

def get_module_hook_type(
    parent_module: Optional[torch.nn.Module],
    cur_module: torch.nn.Module,
) -> HookType:
    cached_hook_type = getattr(cur_module, '_auto_quant_module_hook_type', None)
    if cached_hook_type is not None:
        return cached_hook_type
    parent_module_has_qstate = parent_module is not None and \
        '_auto_quant_state' in parent_module.__dict__
    needs_op_hooks = parent_module_has_qstate and \
        parent_module.__dict__['_auto_quant_state'].cur_op_needs_hooks(cur_module)  # type: ignore[union-attr, operator]
    # We need IO hooks if
    # * we are calling forward on a module (always True here)
    # * that module has quant state
    # * that module does not need op hooks for the parent
    needs_io_hooks = (
        '_auto_quant_state' in cur_module.__dict__ and
        (not needs_op_hooks)
    )
    needs_arg_dequants = parent_module_has_qstate and not needs_op_hooks

    if needs_op_hooks:
        result = HookType.OP_HOOKS
    elif needs_io_hooks:
        result = HookType.MODULE_IO_HOOKS
    elif needs_arg_dequants:
        result = HookType.ARG_DEQUANTS
    else:
        result = HookType.NONE
    cur_module._auto_quant_module_hook_type = result  # type: ignore[assignment]
    return result


def attach_scale_zp_values_to_model(
    module: torch.nn.Module,
) -> None:
    """
    Calculates the scale and zero_point from each observer and attaches
    these values to the parent module. This is done to avoid recalculating
    these values at inference.
    """
    if hasattr(module, '_auto_quant_state'):
        qstate: AutoQuantizationState = module._auto_quant_state  # type: ignore[assignment]
        quantized_dtype =  [torch.quint8, torch.qint8]
        for tensor_id, observer in qstate.tensor_id_to_observer.items():
            if observer.dtype in quantized_dtype:
                scale, zp = observer.calculate_qparams()
                qstate.tensor_id_to_scale_zp[int(tensor_id)] = (scale, zp)
        for tensor_id, observer in qstate.weight_tensor_id_to_observer.items():
            if observer.dtype in quantized_dtype:
                scale, zp = observer.calculate_qparams()
                qstate.weight_tensor_id_to_scale_zp[tensor_id] = (scale, zp)
        qstate.tensor_id_to_observer.clear()
        qstate.weight_tensor_id_to_observer.clear()

    for _, child in module.named_children():
        attach_scale_zp_values_to_model(child)

def attach_op_convert_info_to_model(
    module: torch.nn.Module,
) -> None:
    """
    Calculates the info needed to convert each op and attaches
    it to the parent module. This is done to avoid recalculating these values
    at inference.
    """
    if hasattr(module, '_auto_quant_state'):
        qstate: AutoQuantizationState = module._auto_quant_state  # type: ignore[assignment]
        for _, seen_q_op_info in qstate.idx_to_seen_q_op_infos.items():
            qstate.idx_to_op_convert_info[seen_q_op_info.idx] = \
                qstate.calculate_op_convert_info(seen_q_op_info)
            qstate.idx_to_op_weight_convert_info[seen_q_op_info.idx] = \
                qstate.calculate_op_weight_convert_info(seen_q_op_info)

    for _, child in module.named_children():
        attach_op_convert_info_to_model(child)

class Node:
    def __init__(self, op_infos, input_scale_zero=None, weight_scale_zero=None, output_scale_zero=None, qconfig=None):
        self.idx = op_infos.idx if hasattr(op_infos, 'idx') else None
        self.type = op_infos.type
        self.fqn = op_infos.fqn
        self.input_tensor_infos = op_infos.input_tensor_infos
        self.input_tensor_force_inf_dtype = [] if qconfig is None else op_infos.input_tensor_force_inf_dtype
        self.output_tensor_infos = op_infos.output_tensor_infos
        self.insert_fake_quant_after_outputs = [] if qconfig is None else op_infos.insert_fake_quant_after_outputs
        self.weight_tensor_infos = [] if qconfig is None else op_infos.weight_tensor_infos
        self.qconfig = qconfig
        self.input_scale_zero = input_scale_zero
        self.weight_scale_zero = weight_scale_zero
        self.output_scale_zero = output_scale_zero
        self.pre_nodes = []
        self.post_nodes = []

class ParentNode:
    def __init__(self, output_info):
        self.output_info = output_info

def convert_quant_state_map_to_nodes(quant_state_map):
    nodes = []
    # step1: create nodes
    for _, v in quant_state_map.items():
        new_parent_node = ParentNode(v.output_qtensor_infos)
        nodes.append(new_parent_node)
        for idx, q_op_info in v.idx_to_seen_q_op_infos.items():
            input_scale_zero = {}
            weight_scale_zero = {}
            output_scale_zero = {}
            for tensor_info in q_op_info.input_tensor_infos:
                if tensor_info is not None and tensor_info.id in v.tensor_id_to_scale_zp:
                    input_scale_zero[tensor_info.id] = v.tensor_id_to_scale_zp[tensor_info.id]
            for tensor_info in q_op_info.weight_tensor_infos:
                if tensor_info is not None and (str(idx) + "_" + str(tensor_info.id) in v.weight_tensor_id_to_scale_zp):
                    weight_id = str(idx) + "_" + str(tensor_info.id)
                    weight_scale_zero[weight_id] = v.weight_tensor_id_to_scale_zp[weight_id]
            for tensor_info in q_op_info.output_tensor_infos:
                if tensor_info is not None and tensor_info.id in v.tensor_id_to_scale_zp:
                    output_scale_zero[tensor_info.id] = v.tensor_id_to_scale_zp[tensor_info.id]
            new_node = Node(q_op_info, input_scale_zero, weight_scale_zero, output_scale_zero, qconfig=q_op_info.qconfig)
            nodes.append(new_node)
        for nonq_op_infos in v.seen_nonq_op_infos:
            new_node = Node(nonq_op_infos)
            nodes.append(new_node)
    # create connection between nodess
    for cur in nodes:
        if isinstance(cur, ParentNode):
            continue
        for n in nodes:
            if isinstance(n, ParentNode):
                continue
            # find pre_node:
            for input_info in cur.input_tensor_infos:
                if input_info in n.output_tensor_infos:
                    cur.pre_nodes.append(n)
            # find nex_node:
            for output_info in cur.output_tensor_infos:
                if output_info in n.input_tensor_infos:
                    cur.post_nodes.append(n)
    return nodes

def sync_pool_and_lstm_input_output_scale_zp(quant_state_map, nodes):
    pool_ops = [str(F.adaptive_avg_pool2d), str(F.adaptive_avg_pool3d), str(F.avg_pool2d), str(F.avg_pool3d), \
        str(F.max_pool2d), str(F.max_pool3d), str(nn.MaxPool2d), str(nn.MaxPool3d), str(nn.AvgPool2d), str(nn.AvgPool3d), \
        str(nn.AdaptiveAvgPool2d), str(nn.AdaptiveAvgPool3d)]
    shape_ops = [str(torch.flatten), str(torch.nn.Flatten), str(torch.Tensor.flatten)]
    rnn_ops = [str(torch.nn.LSTM)]
    def _sync_scale_zp_given_id(quant_state_map, id, scale_zp):
        for _, v in quant_state_map.items():
            if id in v.tensor_id_to_scale_zp:
                v.tensor_id_to_scale_zp[id] = scale_zp

    def _find_sync_op_from_given_node(cur_node, ids):
        for next in cur_node.post_nodes:
            if next.type in (pool_ops + shape_ops + rnn_ops):
                ids.append(next.output_tensor_infos[0].id)
                _find_sync_op_from_given_node(next, ids)

    for node in nodes:
        if isinstance(node, ParentNode):
            continue
        if node.qconfig is not None and node.type in (pool_ops + rnn_ops):
            if node.input_scale_zero == node.output_scale_zero:
                continue
            sync_node_begin = node
            # fistly, find the fist sync op before the cur pooling(lstm) op.
            # like: pooling->pool->shape->cur_node,
            while len(sync_node_begin.pre_nodes) == 1 and \
                sync_node_begin.pre_nodes[0].type in (pool_ops + shape_ops + rnn_ops):
                sync_node_begin = sync_node_begin.pre_nodes[0]
            tensor_ids = [sync_node_begin.output_tensor_infos[0].id]
            scale_zp = sync_node_begin.input_scale_zero[sync_node_begin.input_tensor_infos[0].id]
            # find all need sync op from sync_node_begin.
            _find_sync_op_from_given_node(sync_node_begin, tensor_ids)
            for id in tensor_ids:
                _sync_scale_zp_given_id(quant_state_map, id, scale_zp)


def _check_after_nodes_all_quantized_give_node(node):
    r"""
    This function is about check whether given node's post nodes are all quantized,
    """
    if len(node.post_nodes) == 0:
        return False
    else:
        # make sure all post nodes are quantizabled. 
        for next in node.post_nodes:
            if next.qconfig is None:
                if next.type == str(nn.Identity):
                    return _check_after_nodes_all_quantized_give_node(next)
                else:
                    return False
            else:
                int8_int8_symmetric_ops = [str(interaction), str(torch.ops.torch_ipex.interaction_forward), str(torch.embedding_bag), \
                    str(F.embedding_bag), str(torch.nn.EmbeddingBag)]
                if next.type in int8_int8_symmetric_ops:
                    if next.type in [str(interaction), str(torch.ops.torch_ipex.interaction_forward)]:
                        # node.input_tensor_infos may be set, we can use force_inf_dtype to check whether this op is quantizabled.
                        for force_inf_dtype in next.input_tensor_force_inf_dtype:
                            if force_inf_dtype != torch.qint8:
                                return False
                    else:
                        # embeddingBag
                        if next.weight_tensor_infos[0].inf_dtype != torch.qint8:
                            return False
                else:
                    for force_inf_dtype in next.input_tensor_force_inf_dtype:
                        if force_inf_dtype not in [torch.qint8, torch.quint8]:
                            return False
        # all post nodes are quantizabled.
        return True

def set_node_output_quantized(nodes):
    r"""
    # For interation EmbeddingBag, pooling and elt_wise, we only support INT8->INT*, if those ops have quantized their inputs,
    # we need make sure their output also have falk quant to make them call in INT8 kernel.
    # this function will check whether the output inf dtype is int8 dtype if its' input is set to quantized, if the
    # output's infe dtype is not int8, set it and also set insert_fake_quant_after_output to True.
    """
    def _reset_post_node_input_infos(node):
        # make sure the post node will node insert fake quant if we add fake quant by cur node' output
        if len(node.post_nodes) > 0:
            for post_node in node.post_nodes:
                if post_node.qconfig is not None:
                    for idx, tensor_info in enumerate(post_node.input_tensor_infos):
                        if tensor_info in node.output_tensor_infos:
                            post_node.input_tensor_force_inf_dtype[idx] = tensor_info.orig_dtype
                elif post_node.type == str(nn.Identity):
                    _reset_post_node_input_infos(post_node)

    for node in nodes:
        if isinstance(node, ParentNode):
            continue
        if node.qconfig is not None and node.type in int8_int8_ops:
            post_node_are_quantized = _check_after_nodes_all_quantized_give_node(node)
            if node.type in str(torch.nn.EmbeddingBag):
                if node.weight_tensor_infos[0].inf_dtype == torch.qint8 and not post_node_are_quantized:
                    node.output_tensor_infos[0].inf_dtype = torch.qint8
                    node.insert_fake_quant_after_outputs[0] = True
                    _reset_post_node_input_infos(node)
            elif node.type  == str(F.embedding_bag):
                if node.input_tensor_force_inf_dtype[1] == torch.qint8 and not post_node_are_quantized:
                    node.output_tensor_infos[0].inf_dtype = torch.qint8
                    node.insert_fake_quant_after_outputs[0] = True
                    _reset_post_node_input_infos(node)
            elif node.type in [str(interaction), str(torch.ops.torch_ipex.interaction_forward)]:
                if node.input_tensor_force_inf_dtype[0] == torch.qint8 and not post_node_are_quantized:
                    node.output_tensor_infos[0].inf_dtype = torch.qint8
                    node.insert_fake_quant_after_outputs[0] = True
                    _reset_post_node_input_infos(node)
            else:
                if node.input_tensor_force_inf_dtype[0] in [torch.qint8, torch.quint8] and not post_node_are_quantized:
                   node.output_tensor_infos[0].inf_dtype = node.input_tensor_force_inf_dtype[0]
                   node.insert_fake_quant_after_outputs[0] = True
                   _reset_post_node_input_infos(node)

qscheme_dict = {
    str(torch.per_tensor_affine): torch.per_tensor_affine,
    str(torch.per_tensor_symmetric): torch.per_tensor_symmetric,
    str(torch.per_channel_affine): torch.per_channel_affine,
    str(torch.per_channel_symmetric): torch.per_channel_symmetric
}

dtype_dict = {
    str(torch.quint8): torch.quint8,
    str(torch.qint8): torch.qint8,
    str(torch.float32): torch.float32,
    str(torch.float64): torch.float64,
    str(torch.float16): torch.float16,
    str(torch.bfloat16): torch.bfloat16,
    str(torch.complex64): torch.complex64,
    str(torch.complex128): torch.complex128,
    str(torch.int16): torch.int16,
    str(torch.int32): torch.int32,
    str(torch.int64): torch.int64,
    str(torch.bool): torch.bool,
    str(torch.uint8): torch.uint8,
    str(torch.int8): torch.int8,
    str(torch.quint4x2): torch.quint4x2,
}

def _get_observer_setting(observer):
    r"""
    Convert torch observer's args to dict for saving to json file.
    Because json only accept number or string, so we will convert some
    class type to string(dtype, qscheme, other class type?).
    """
    observer_setting = OrderedDict()
    observer_setting["name"] =  observer.__class__.__name__
    # get observer arg name
    observer_args = inspect.signature(observer.__init__).parameters
    observer_dict = observer.__dict__
    # Now we only can save number or string to json file.
    for arg_name in observer_args.keys():
        if arg_name in observer_dict:
            if isinstance(observer_dict[arg_name], numbers.Number):
                observer_setting[arg_name] = observer_dict[arg_name]
            elif arg_name == "qscheme" or arg_name == "dtype":
                observer_setting[arg_name] = str(observer_dict[arg_name])
            elif arg_name == "eps" and hasattr(observer, "eps") and \
                isinstance(observer.eps, torch.Tensor) and observer.eps.numel() == 1:
                observer_setting[arg_name] = observer.eps.item()
    return observer_setting

def _create_observer(setting):
    r"""
    Create torch observer according to user's setting.
    """
    if "qscheme" in setting:
        setting["qscheme"] = qscheme_dict[setting["qscheme"]]
    if "dtype" in setting:
        setting["dtype"] = dtype_dict[setting["dtype"]]

    if hasattr(torch.quantization.observer, setting["name"]):
        observer = getattr(torch.quantization.observer, setting["name"])
        setting.pop("name", None)
        return observer.with_args(**setting)
    else:
        raise NameError('torch.quantization.observer %s not found' % setting["name"])

def save_quant_state(quant_state_map, configure_file):
    # save qparam's as json file for tunning
    quant_state_dict = OrderedDict()
    for k, v in quant_state_map.items():
        layer_infos = OrderedDict()
        if len(v.idx_to_seen_q_op_infos) == 0:
            layer_infos["q_op_infos"] = {}
        else:
            q_op_infos = OrderedDict()
            for q_k, op_info in v.idx_to_seen_q_op_infos.items():
                info = OrderedDict()
                info["op_type"] = op_info.type
                info["op_type_is_module"] = op_info.type_is_module
                info["fqn"] = op_info.fqn
                input_tensor_infos = []
                for tensor_info, force_dtype in zip(op_info.input_tensor_infos, op_info.input_tensor_force_inf_dtype):
                    cur_tensor_infos = {}

                    if tensor_info is not None:
                        cur_tensor_infos["id"] = tensor_info.id
                        cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                        cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                        cur_tensor_infos["force_dtype"] = str(force_dtype)
                        if tensor_info.id in v.tensor_id_to_scale_zp:
                            cur_tensor_infos["scale"] = v.tensor_id_to_scale_zp[tensor_info.id][0].tolist()
                            cur_tensor_infos["zero_point"] = v.tensor_id_to_scale_zp[tensor_info.id][1].tolist()
                    input_tensor_infos.append(cur_tensor_infos)
                info["input_tensor_infos"] = input_tensor_infos
                # weight infos
                weight_tensor_infos = []
                for tensor_info in op_info.weight_tensor_infos:
                    cur_tensor_infos = {}
                    if tensor_info is not None:
                        cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                        cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                        weight_idx = str(op_info.idx) + "_" + str(tensor_info.id)
                        if weight_idx in v.weight_tensor_id_to_scale_zp:
                            cur_tensor_infos["scale"] = v.weight_tensor_id_to_scale_zp[weight_idx][0].tolist()
                            cur_tensor_infos["zero_point"] = v.weight_tensor_id_to_scale_zp[weight_idx][1].tolist()
                    weight_tensor_infos.append(cur_tensor_infos)
                info["weight_tensor_infos"] = weight_tensor_infos
                # output infos
                output_tensor_infos = []
                for tensor_info in op_info.output_tensor_infos:
                    cur_tensor_infos = {}
                    if tensor_info is not None:
                        cur_tensor_infos["id"] = tensor_info.id
                        cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                        cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                        if tensor_info.id in v.tensor_id_to_scale_zp:
                            cur_tensor_infos["scale"] = v.tensor_id_to_scale_zp[tensor_info.id][0].tolist()
                            cur_tensor_infos["zero_point"] = v.tensor_id_to_scale_zp[tensor_info.id][1].tolist()
                    output_tensor_infos.append(cur_tensor_infos)
                info["output_tensor_infos"] = output_tensor_infos
                # qconfig
                info["activation_observer"] = _get_observer_setting(op_info.qconfig.activation())
                info["weight_observer"] = _get_observer_setting(op_info.qconfig.weight())
                q_op_infos[q_k] = info
            layer_infos["q_op_infos"] = q_op_infos
        if len(v.seen_nonq_op_infos) == 0:
            layer_infos["nonq_op_infos"] = {}
        else:
            nonq_op_infos = OrderedDict()
            for non_q_k, op_info in enumerate(v.seen_nonq_op_infos):
                info = OrderedDict()
                info["op_type"] = op_info.type
                info["fqn"] = op_info.fqn
                input_tensor_infos = []
                for tensor_info in op_info.input_tensor_infos:
                     cur_tensor_infos = {}
                     if tensor_info is not None:
                        cur_tensor_infos["id"] = tensor_info.id
                        cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                        cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                     input_tensor_infos.append(cur_tensor_infos)
                info["input_tensor_infos"] = input_tensor_infos
                output_tensor_infos = []
                for tensor_info in op_info.output_tensor_infos:
                     cur_tensor_infos = {}
                     if tensor_info is not None:
                        cur_tensor_infos["id"] = tensor_info.id
                        cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                        cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                     output_tensor_infos.append(cur_tensor_infos)
                info["output_tensor_infos"] = output_tensor_infos
                nonq_op_infos[non_q_k] = info
            layer_infos["nonq_op_infos"] = nonq_op_infos
        layer_output_infos = []
        for tensor_info in v.output_qtensor_infos:
            cur_tensor_infos = {}
            if tensor_info is not None:
                cur_tensor_infos["id"] = tensor_info.id
                cur_tensor_infos["orig_dtype"] = str(tensor_info.orig_dtype)
                cur_tensor_infos["inf_dtype"] = str(tensor_info.inf_dtype)
                if tensor_info.id in v.tensor_id_to_scale_zp:
                    cur_tensor_infos["scale"] = v.tensor_id_to_scale_zp[tensor_info.id][0].tolist()
                    cur_tensor_infos["zero_point"] = v.tensor_id_to_scale_zp[tensor_info.id][1].tolist()
            layer_output_infos.append(cur_tensor_infos)
        layer_infos["layer_output_infos"] = layer_output_infos
        quant_state_dict[k] = layer_infos
    # save qparms as json file
    if configure_file is not None:
        with open(configure_file, 'w') as fp:
            json.dump(quant_state_dict, fp, indent = 4)

def load_qconf_summary_to_model(model, qconf_summary):
    """
    This function is about load the user given configure to origin model.
    """
    with open(qconf_summary, 'r') as f:
        quant_state_dict = json.load(f)
    quant_state_map = model._fqn_to_auto_quant_state_map
    for k, v in quant_state_map.items():
        layer_info = quant_state_dict[k]
        user_q_op_infos = layer_info["q_op_infos"]
        for i, q_op_info in user_q_op_infos.items():
            fqn = q_op_info["fqn"]
            cur_fqn = v.idx_to_seen_q_op_infos[int(i)].fqn
            assert int(i) in v.idx_to_seen_q_op_infos and \
                    cur_fqn == fqn, "Loaded quantizable op info doesn't match the current model quantizable op info"
            input_tensor_infos = []
            input_force_dtype_infos = []
            for tensor_info in q_op_info["input_tensor_infos"]:
                if len(tensor_info) > 0:
                    input_tensor_infos.append(QTensorInfo(tensor_info["id"], dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                    input_force_dtype_infos.append(dtype_dict[tensor_info["force_dtype"]])
                    if "scale" in tensor_info:
                        scale = torch.FloatTensor(tensor_info["scale"])
                        zp = torch.LongTensor(tensor_info["zero_point"])
                        v.tensor_id_to_scale_zp[tensor_info["id"]] = (scale, zp)
                else:
                    input_tensor_infos.append(None)
            weight_tensor_infos = []
            weight_idx = 0
            for tensor_info in q_op_info["weight_tensor_infos"]:
                if len(tensor_info) > 0:
                    weight_tensor_infos.append(QTensorInfo(weight_idx, dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                    if "scale" in tensor_info:
                        scale = torch.FloatTensor(tensor_info["scale"])
                        zp = torch.LongTensor(tensor_info["zero_point"])
                        v.weight_tensor_id_to_scale_zp[str(i) + "_" + str(weight_idx)] = (scale, zp)
                    weight_idx += 1
                else:
                    weight_tensor_infos.append(None)
            output_tensor_infos = []
            insert_fake_quant_after_outputs = []
            for tensor_info in q_op_info["output_tensor_infos"]:
                if len(tensor_info) > 0:
                    output_tensor_infos.append(QTensorInfo(tensor_info["id"], dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                    insert_fake_quant_after_outputs.append(False)
                    if "scale" in tensor_info:
                        scale = torch.FloatTensor(tensor_info["scale"])
                        zp = torch.LongTensor(tensor_info["zero_point"])
                        v.tensor_id_to_scale_zp[tensor_info["id"]] = (scale, zp)
                else:
                    output_tensor_infos.append(None)
            activation_observer = q_op_info["activation_observer"]
            weight_observer = q_op_info["weight_observer"]
            qconfig = QConfig(activation=_create_observer(activation_observer), weight = _create_observer(weight_observer))
            # overide the cur model's info
            v.idx_to_seen_q_op_infos[int(i)].input_tensor_infos = input_tensor_infos
            v.idx_to_seen_q_op_infos[int(i)].input_tensor_force_inf_dtype = input_force_dtype_infos
            v.idx_to_seen_q_op_infos[int(i)].output_tensor_infos = output_tensor_infos
            v.idx_to_seen_q_op_infos[int(i)].insert_fake_quant_after_outputs = insert_fake_quant_after_outputs
            v.idx_to_seen_q_op_infos[int(i)].weight_tensor_infos = weight_tensor_infos
            v.idx_to_seen_q_op_infos[int(i)].qconfig = qconfig

        user_nonq_op_infos = layer_info["nonq_op_infos"]
        #v.seen_nonq_op_infos.clear()
        idx = 0
        for _, nonq_op_info in user_nonq_op_infos.items():
            fqn = nonq_op_info["fqn"]
            cur_fqn = v.seen_nonq_op_infos[idx].fqn
            assert cur_fqn == fqn, "Loaded none-quantizable op info doesn't match the current model none-quantizable op info"
            input_tensor_infos = []
            for tensor_info in nonq_op_info["input_tensor_infos"]:
                if len(tensor_info) > 0:
                    input_tensor_infos.append(QTensorInfo(tensor_info["id"], dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                else:
                    input_tensor_infos.append(None)
            output_tensor_infos = []
            for tensor_info in nonq_op_info["output_tensor_infos"]:
                if len(tensor_info) > 0:
                    output_tensor_infos.append(QTensorInfo(tensor_info["id"], dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                else:
                    output_tensor_infos.append(None)
            v.seen_nonq_op_infos[idx].input_tensor_infos = input_tensor_infos
            v.seen_nonq_op_infos[idx].output_tensor_infos = output_tensor_infos
            idx += 1

        layer_output_info = []
        for tensor_info in layer_info["layer_output_infos"]:
            if len(tensor_info) > 0:
                layer_output_info.append(QTensorInfo(tensor_info["id"], dtype_dict[tensor_info["orig_dtype"]], dtype_dict[tensor_info["inf_dtype"]]))
                if "scale" in tensor_info:
                    scale = torch.FloatTensor(tensor_info["scale"])
                    zp = torch.LongTensor(tensor_info["zero_point"])
                    v.tensor_id_to_scale_zp[tensor_info["id"]] = (scale, zp)
            else:
                layer_output_info.append(None)
        v.output_qtensor_infos = layer_output_info
    # insert observer according to user's setting.
    for _, v in model.named_modules():
        if hasattr(v, '_auto_quant_state'):
            v._auto_quant_state.tensor_id_to_observer.clear()
            v._auto_quant_state.weight_tensor_id_to_observer.clear()
            v._auto_quant_state.insert_observers(v)
    
    # update insert_fake_quant_after_output after load user setting, which for avoiding redundant fake quant.
    nodes = convert_quant_state_map_to_nodes(quant_state_map)
    set_node_output_quantized(nodes)

def _lstm_forward(module, input, hx, weights):
    r"""
    LSTM forward function.
    """
    orig_input = input
    # xxx: isinstance check needs to be in conditional for TorchScript to compile
    # batch_sizes = None
    if isinstance(orig_input, torch.nn.utils.rnn.PackedSequence):
        input, batch_sizes, sorted_indices, unsorted_indices = input
        max_batch_size = batch_sizes[0]
        max_batch_size = int(max_batch_size)
    else:
        batch_sizes = None
        is_batched = input.dim() == 3
        batch_dim = 0 if module.batch_first else 1
        if not is_batched:
            input = input.unsqueeze(batch_dim)
        max_batch_size = input.size(0) if module.batch_first else input.size(1)
        sorted_indices = None
        unsorted_indices = None
    if hx is None:
        num_directions = 2 if module.bidirectional else 1
        real_hidden_size = module.proj_size if module.proj_size > 0 else module.hidden_size
        h_zeros = torch.zeros(module.num_layers * num_directions, max_batch_size, real_hidden_size, \
            dtype=input.dtype, device=input.device)
        c_zeros = torch.zeros(module.num_layers * num_directions, max_batch_size, module.hidden_size, \
            dtype=input.dtype, device=input.device)
        hx = (h_zeros, c_zeros)
    else:
        if batch_sizes is None:  # If not PackedSequence input.
            if is_batched:
                if (hx[0].dim() != 3 or hx[1].dim() != 3):
                    msg = ("For batched 3-D input, hx and cx should "
                            f"also be 3-D but got ({hx[0].dim()}-D, {hx[1].dim()}-D) tensors")
                    raise RuntimeError(msg)
            else:
                if hx[0].dim() != 2 or hx[1].dim() != 2:
                    msg = ("For unbatched 2-D input, hx and cx should "
                               f"also be 2-D but got ({hx[0].dim()}-D, {hx[1].dim()}-D) tensors")
                    raise RuntimeError(msg)
                hx = (hx[0].unsqueeze(1), hx[1].unsqueeze(1))

        # Each batch of the hidden state should match the input sequence that
        # the user believes he/she is passing in.
        hx = module.permute_hidden(hx, sorted_indices)

    module.check_forward_args(input, hx, batch_sizes)
    if batch_sizes is None:
        result = _VF.lstm(input, hx, weights, module.bias, module.num_layers, \
            module.dropout, module.training, module.bidirectional, module.batch_first)
    else:
        result = _VF.lstm(input, batch_sizes, hx, weights, module.bias, \
            module.num_layers, module.dropout, module.training, module.bidirectional)
    output = result[0]
    hidden = result[1:]
    # xxx: isinstance check needs to be in conditional for TorchScript to compile
    if isinstance(orig_input, torch.nn.utils.rnn.PackedSequence):
        output_packed = torch.nn.utils.rnn.PackedSequence(output, batch_sizes, sorted_indices, unsorted_indices)
        return output_packed, module.permute_hidden(hidden, unsorted_indices)
    else:
        if not is_batched:
            output = output.squeeze(batch_dim)
            hidden = (hidden[0].squeeze(1), hidden[1].squeeze(1))
        return output, module.permute_hidden(hidden, unsorted_indices)

def module_call_to_function_call(module, args, weights):
    r"""
    This function is a help function which replace nn.module call to funtion call, which implement
    the nn.module's forward function.
    """
    if isinstance(module, torch.nn.Conv2d) or isinstance(module, torch.nn.Conv3d):
        output = module._conv_forward(args[0], weights[0], module.bias)
    elif isinstance(module, torch.nn.Linear):
        output = F.linear(args[0], weights[0], module.bias)
    elif isinstance(module, torch.nn.EmbeddingBag):
        output = F.embedding_bag(args[0], weights[0], args[1], module.max_norm, \
            module.norm_type, module.scale_grad_by_freq, module.mode, module.sparse,
            args[2] if len(args) == 3 else None, module.include_last_offset, module.padding_idx)
    elif isinstance(module, torch.nn.ConvTranspose2d) or isinstance(module, torch.nn.ConvTranspose2d):
        if module.padding_mode != 'zeros':
            raise ValueError('Only `zeros` padding mode is supported for ConvTranspose2d')
        assert isinstance(module.padding, tuple)
        # One cannot replace List by Tuple or Sequence in "_output_padding" because
        # TorchScript does not support `Sequence[T]` or `Tuple[T, ...]`.
        output_size = args[1] if len(args) == 2 else None
        # master code
        '''
        num_spatial_dims = 2 if isinstance(module, torch.nn.ConvTranspose2d) else 3
        output_padding = module._output_padding(args[0], output_size,
                        module.stride, module.padding, module.kernel_size,
                        num_spatial_dims, module.dilation)
        '''
        output_padding = module._output_padding(args[0], output_size, module.stride, module.padding, module.kernel_size,  module.dilation)
        #output_padding = module._output_padding(*arg_to)
        if isinstance(module, torch.nn.ConvTranspose2d):
            output = F.conv_transpose2d(
                args[0], weights[0], module.bias, module.stride, module.padding,
                output_padding, module.groups, module.dilation)
        else:
            output = F.conv_transpose3d(
                args[0], weights[0], module.bias, module.stride, module.padding,
                output_padding, module.groups, module.dilation)
    elif isinstance(module, torch.nn.LSTM):
        output = _lstm_forward(module, args[0], args[1] if len(args) == 2 else None, weights)
    return output



import torch
import copy
import types
import warnings
from ._functional import sgd_step, adagrad_step, lamb_step, adam_step
from ._lamb import Lamb
from ..nn import utils

IPEX_FUSED_OPTIMIZER_LIST = [
    torch.optim.SGD,
    torch.optim.Adagrad,
    torch.optim.Adam,
    Lamb,
]

OPTIMIZER_FUSED_STEP_MAPPING = {
    torch.optim.SGD: sgd_step,
    torch.optim.Adagrad: adagrad_step,
    torch.optim.Adam: adam_step,
    Lamb: lamb_step
}

def patch_zero_grad_for_master_weight_training(optimizer):
    r"""
    Patch "zero_grad" method of optimizer to support BFloat16 master weight training
    Under master weight training case, the grad is actually on 'bf16_params'. So the 'zero_grad'
    should work on the 'bf16_params' too.
    """
    def zero_grad(self, set_to_none: bool = False):
        for p in self.params_attr:
            if 'bf16_param' in self.params_attr[p]:
                bf16_param = self.params_attr[p]['bf16_param']
            if bf16_param.grad is not None:
                if set_to_none:
                    bf16_param.grad = None
                else:
                    if bf16_param.grad.grad_fn is not None:
                        bf16_param.grad.detach_()
                    else:
                        bf16_param.grad.requires_grad_(False)
                    bf16_param.grad.zero_()
        self._original_zero_grad(set_to_none)
    setattr(optimizer, '_original_zero_grad', optimizer.zero_grad)
    setattr(optimizer, 'zero_grad', types.MethodType(zero_grad, optimizer))

def patch_step_for_master_weight_training(optimizer):
    r"""
    Patch "step" method of optimizer to support BFloat16 master weight training
    1.Convert BF16 grad to FP32
    2.Call original "step" to update parameters
    3.Sync FP32 master weight back to BF16 weight
    """
    def master_param_non_fused_step(self, closure=None):
        # convert bf16 weight'grad to float.
        for k, value in self.params_attr.items():
           if value['bf16_param'].requires_grad:
                k.grad = value['bf16_param'].grad.detach().float()

        loss = self._original_step(closure)
        # sync mater weight to model's paramerter
        for k, value in self.params_attr.items():
            torch.ops.torch_ipex.sync_master_weight_to_bf16(k, value['bf16_param'])
        return loss
    setattr(optimizer, '_original_step', optimizer.step)
    setattr(optimizer, 'step', types.MethodType(master_param_non_fused_step, optimizer))

def patch_load_state_dict(optimizer):
    r"""
    Forbid optimizer load state dict after weight-prepack or weight-cast
    """
    def load_state_dict(self, state_dict):
        assert False, "_ipex_optimizer does not suppory load_state_dict"
    setattr(optimizer, '_original_load_state_dict', optimizer.load_state_dict)
    setattr(optimizer, 'load_state_dict', types.MethodType(load_state_dict, optimizer))

def refresh_optimizer_params_after_cast(m, attr, float_param, master_weight_split, optimizer):
    r"""
    After casting nn.Modules parameters, need refresh corresponding parameters in optimizers
    For master weight solution, the parameters in optimizers should be master weight (not BF16 weight)
    """
    if optimizer is None:
        return
    # update params
    for group in optimizer.param_groups:
        for i, p in enumerate(group['params']):
            if p is float_param:
                if master_weight_split:
                    group['params'][i] = getattr(m, attr)
                else:
                    group['params'][i] = getattr(m, 'master_' + attr)
                # update optimizer's state.
                new_param = group['params'][i]
                if p in optimizer.state:
                    optimizer.state[new_param] = optimizer.state.pop(p)

def pack_optimizer_params_and_states(optimizer, param_pair, attrs, pack_dtype):
    """
    1. convert user's optimizer weights and related states to packed format
    While optimizer is maintain "master weight", the key in attrs is "weight",
    Need pass "weight" here as attr_key visit attr.
    2. convert user's optimizer bias to new model's bias since there is a "clone"
    """
    if optimizer is None:
        return
    for group in optimizer.param_groups:
        for i, p in enumerate(group['params']):
            if p in param_pair:
                new_param = param_pair[p]
                group['params'][i] = new_param
                # copy optimizer's state.
                if p in optimizer.state:
                    optimizer.state[new_param] = optimizer.state.pop(p)
                    # Prepack the state according to the prepacked weight.
                    # it covers both conv and linear now. TODO: LSTM or other ops.
                    if new_param in attrs:
                        attr = attrs[new_param]
                        if 'op' in attr:
                            # weight attr need "op" info to pack state while bias attr not
                            state = optimizer.state[new_param]
                            for state_key, state_value in state.items():
                                if isinstance(state_value, torch.Tensor) and state_value.size() == p.size():
                                    # We have an assumtion here that any tensor's in parameter state, if they
                                    # have same shapes with the parameter, they should share same layout with 
                                    # the parameter. Thus we need pack the state as we did to parameters.
                                    if attr['op'] in utils._weight_prepack.IPEX_WEIGHT_PREPACK_MODULE:
                                        if attr['op'] is torch.nn.Conv1d or attr['op'] is torch.nn.Conv2d or attr['op'] is torch.nn.Conv3d:
                                            if attr['op'] is torch.nn.Conv2d:
                                                memory_format = torch.channels_last
                                            elif attr['op'] is torch.nn.Conv3d:
                                                memory_format = torch.channels_last_3d
                                            else:
                                                memory_format = torch.contiguous_format
                                            value_temp = state_value.to(memory_format=memory_format) \
                                                if attr['weight_channels_last'] else state_value
                                            state[state_key] = attr['ctx'].pack(value_temp)
                                        else:
                                            state[state_key] = attr['ctx'].pack(state_value)                                   

def patch_state_dict(optimizer):
    r"""
    To support resume training.
    Patch "state_dict" method to return unpacked/FP32 parameters/states
    """
    def get_optimizer_unpacked_state_dict(self):
        opt = self
        opt_temp = copy.deepcopy(opt)
        for (k1, _), (_, v2) in zip(opt.state.items(), opt_temp.state.items()):
            if k1 in opt.params_attr:
                params_attr = opt.params_attr[k1]
                for state_key, state_value in v2.items():
                    if isinstance(state_value, torch.Tensor) and state_value.shape == k1.shape:
                        # We have an assumtion here that any tensor's in parameter state, if they
                        # have same shapes with the parameter, they should share same layout with 
                        # the parameter. Thus we need unpack the state as we did to parameters.
                        if 'op' in params_attr:
                            # Secondly, unpack releated states
                            if params_attr['op'] in utils._weight_prepack.IPEX_WEIGHT_PREPACK_MODULE:
                                state_value = params_attr['ctx'].to_public(state_value)
                            else:
                                assert False, "unsupported op to unpack"
                        v2[state_key] = state_value
        return opt_temp.state_dict()
    setattr(optimizer, '_original_state_dict', optimizer.state_dict)
    setattr(optimizer, 'state_dict', types.MethodType(get_optimizer_unpacked_state_dict, optimizer))

def optimizer_fusion(optimizer, master_weight_split):
    r"""
    Patch "step" method to choose IPEX optimized fused update kernel.
    """
    setattr(optimizer, 'fused', True)
    if not hasattr(optimizer, 'params_attr'):
        setattr(optimizer, 'params_attr', {})
    try:
        step = OPTIMIZER_FUSED_STEP_MAPPING[type(optimizer)]
        if not hasattr(optimizer, '_original_step'):
            setattr(optimizer, '_original_step', optimizer.step)
        setattr(optimizer, 'step', types.MethodType(step, optimizer))
    except KeyError:
        warnings.warn("Does not suport fused step for " + str(type(optimizer)) + ", will use non-fused step")
    return optimizer



r"""Functional interface, port from torch/optim/_function.py"""
import torch
from torch import Tensor
from typing import List, Optional

def is_master_weight(param, params_attr):
    return  (
        param.dtype == torch.float and
        param in params_attr and
        'bf16_param' in params_attr[param]
    )

def get_bf16_grad(param, params_attr):
    assert is_master_weight(param, params_attr)
    return params_attr[param]['bf16_param'].grad

def get_param2(param, params_attr):
    # For pure fp32 case, param2 is not needed.
    # For master weight case, param2 is the bf16 copy of fp32 weight
    # For master weight split case, param2 is the trail part of fp32 weight
    param2 = torch.Tensor()
    if param in params_attr:
        if 'trail' in params_attr[param]:
            assert param.dtype is torch.bfloat16
            param2 = params_attr[param]['trail']
        if 'bf16_param' in params_attr[param]:
            assert param.dtype is torch.float
            param2 = params_attr[param]['bf16_param']
    return param2

def _make_sparse(grad, grad_indices, values):
    size = grad.size()
    if grad_indices.numel() == 0 or values.numel() == 0:
        return torch.empty_like(grad)
    return torch.sparse_coo_tensor(grad_indices, values, size)

def _single_tensor_adagrad(params: List[Tensor],
                           params2: List[Tensor],
                           grads: List[Tensor],
                           state_sums: List[Tensor],
                           state_steps: List[Tensor],
                           *,
                           lr: float,
                           weight_decay: float,
                           lr_decay: float,
                           eps: float,
                           has_sparse_grad: bool,
                           maximize: bool,
                           fused: bool):

    for (param, param2, grad, state_sum, step_t) in zip(params, params2, grads, state_sums, state_steps):
        # update step
        step_t += 1
        step = step_t.item()
        grad = grad if not maximize else -grad
        if not (grad.is_sparse or torch.is_complex(param)):
            torch.ops.torch_ipex.adagrad_fused_step(
                param,
                grad,
                state_sum,
                param2,
                step,
                lr,
                weight_decay,
                lr_decay,
                eps)
            continue

        if weight_decay != 0:
            if grad.is_sparse:
                raise RuntimeError("weight_decay option is not compatible with sparse gradients")
            grad = grad.add(param, alpha=weight_decay)

        grad = grad.to(param.dtype)
        clr = lr / (1 + (step - 1) * lr_decay)

        if grad.is_sparse:
            grad = grad.coalesce()  # the update is non-linear so indices must be unique
            grad_indices = grad._indices()
            grad_values = grad._values()
            size = grad.size()

            state_sum.add_(_make_sparse(grad, grad_indices, grad_values.pow(2)))
            std = state_sum.sparse_mask(grad)
            std_values = std._values().sqrt_().add_(eps)
            param.add_(_make_sparse(grad, grad_indices, grad_values / std_values), alpha=-clr)
        else:
            is_complex = torch.is_complex(param)
            if is_complex:
                grad = torch.view_as_real(grad)
                state_sum = torch.view_as_real(state_sum)
                param = torch.view_as_real(param)
            state_sum.addcmul_(grad, grad, value=1)
            std = state_sum.sqrt().add_(eps)
            param.addcdiv_(grad, std, value=-clr)
            if is_complex:
                param = torch.view_as_complex(param)
                state_sum = torch.view_as_complex(state_sum)

# keep this function here if enable fused_foreach_adagrad_later
def _multi_tensor_adagrad(params: List[Tensor],
                          params2: List[Tensor],
                          grads: List[Tensor],
                          state_sums: List[Tensor],
                          state_steps: List[Tensor],
                          *,
                          lr: float,
                          weight_decay: float,
                          lr_decay: float,
                          eps: float,
                          has_sparse_grad: bool,
                          maximize: bool,
                          fused: bool):

    # Foreach functions will throw errors if given empty lists
    if len(params) == 0:
        return

    if maximize:
        grads = torch._foreach_neg(grads)

    _single_tensor_adagrad(params,
                            params2,
                            grads,
                            state_sums,
                            state_steps,
                            lr=lr,
                            weight_decay=weight_decay,
                            lr_decay=lr_decay,
                            eps=eps,
                            has_sparse_grad=has_sparse_grad,
                            maximize=False,
                            fused=fused)
    return

def adagrad(params: List[Tensor],
            params2: List[Tensor],
            grads: List[Tensor],
            state_sums: List[Tensor],
            state_steps: List[Tensor],
            # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
            # setting these as kwargs for now as functional API is compiled by torch/distributed/optim
            has_sparse_grad: bool = None,
            foreach: bool = None,
            *,
            lr: float,
            weight_decay: float,
            lr_decay: float,
            eps: float,
            maximize: bool,
            fused: bool):
    r"""Functional API that performs Adagrad algorithm computation.

    See :class:`~torch.optim.Adagrad` for details.
    """

    if not all([isinstance(t, torch.Tensor) for t in state_steps]):
        raise RuntimeError("API has changed, `state_steps` argument must contain a list of singleton tensors")

    if foreach is None:
        # Placeholder for more complex foreach logic to be added when value is not set
        foreach = False

    if foreach and torch.jit.is_scripting():
        raise RuntimeError('torch.jit.script not supported with foreach optimizers')

    if foreach and not torch.jit.is_scripting():
        func = _multi_tensor_adagrad
    else:
        func = _single_tensor_adagrad

    func(params,
         params2,
         grads,
         state_sums,
         state_steps,
         lr=lr,
         weight_decay=weight_decay,
         lr_decay=lr_decay,
         eps=eps,
         has_sparse_grad=has_sparse_grad,
         maximize=maximize,
         fused=fused)

@torch.no_grad()
def adagrad_step(self, closure=None):
    """Performs a single optimization step.

    Args:
        closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
    """
    loss = None
    if closure is not None:
        with torch.enable_grad():
            loss = closure()

    for group in self.param_groups:
        params_with_grad = []
        params2 = []
        grads = []
        state_sums = []
        state_steps = []

        has_sparse_grad = False
        for p in group['params']:
            grad = get_bf16_grad(p, self.params_attr) if is_master_weight(p, self.params_attr) else p.grad
            if grad is not None:
                if grad.is_sparse:
                    has_sparse_grad = True
                params_with_grad.append(p)
                grads.append(grad)
                state = self.state[p]
                state_sums.append(state['sum'])
                state_steps.append(state['step'])
                param2 = get_param2(p, self.params_attr)
                params2.append(param2)

        adagrad(params_with_grad,
                params2,
                grads,
                state_sums,
                state_steps,
                lr=group['lr'],
                weight_decay=group['weight_decay'],
                lr_decay=group['lr_decay'],
                eps=group['eps'],
                has_sparse_grad=has_sparse_grad,
                foreach=group['foreach'],
                maximize=group["maximize"],
                fused=self.fused)

    return loss

def _sgd_non_fused_micro_step(
    param: Tensor,
    grad: Tensor,
    momentum_buffer: Optional[Tensor],
    momentum: float,
    lr: float,
    weight_decay: float,
    dampening: float,
    nesterov: bool,
):
    if weight_decay != 0:
        grad = grad.add(param, alpha=weight_decay)

    if momentum != 0:
        buf = momentum_buffer

        if buf is None:
            buf = torch.clone(grad).detach()
            momentum_buffer = buf
        else:
            buf.mul_(momentum).add_(grad, alpha=1 - dampening)

        if nesterov:
            grad = grad.add(buf, alpha=momentum)
        else:
            grad = buf

    param.add_(grad, alpha=-lr)
    return momentum_buffer

def _single_tensor_sgd(params: List[Tensor],
                      params2: List[Tensor],
                      grads: List[Tensor],
                      momentum_buffer_list: List[Optional[Tensor]],
                      *,
                      weight_decay: float,
                      momentum: float,
                      lr: float,
                      dampening: float,
                      nesterov: bool,
                      maximize: bool,
                      has_sparse_grad: bool,
                      fused: bool):
    if maximize:
        lr = -lr

    for i, param in enumerate(params):
        if not grads[i].is_sparse:
            momentum_buffer_list[i] = torch.ops.torch_ipex.sgd_fused_step(
                param,
                grads[i],
                momentum_buffer_list[i],
                params2[i],
                momentum,
                lr,
                weight_decay,
                dampening,
                nesterov)
            continue

        if (
            param.dtype == torch.bfloat16 and
            grads[i].is_sparse and
            grads[i].dtype == torch.bfloat16 and
            weight_decay == 0 and
            momentum == 0
        ):
            # packed_add can support sparse tensor
            torch.ops.torch_ipex.packed_add(param, params2[i], grads[i], alpha=-lr)
        else:
            # no special optimize for other non fused case, fall back to naive implementation
            grads[i] = grads[i].to(param.dtype)
            momentum_buffer_list[i] = _sgd_non_fused_micro_step(
                param,
                grads[i],
                momentum_buffer_list[i],
                momentum,
                lr,
                weight_decay,
                dampening,
                nesterov
            )

# keep this function here if enable fused_foreach_sgd_later
def _multi_tensor_sgd(params: List[Tensor],
                      params2: List[Tensor],
                      grads: List[Tensor],
                      momentum_buffer_list: List[Optional[Tensor]],
                      *,
                      weight_decay: float,
                      momentum: float,
                      lr: float,
                      dampening: float,
                      nesterov: bool,
                      maximize: bool,
                      has_sparse_grad: bool,
                      fused: bool):

    if len(params) == 0:
        return

    _single_tensor_sgd(params,
                        params2,
                        grads,
                        momentum_buffer_list,
                        weight_decay=weight_decay,
                        momentum=momentum,
                        lr=lr,
                        dampening=dampening,
                        nesterov=nesterov,
                        maximize=maximize,
                        has_sparse_grad=has_sparse_grad,
                        fused=fused)

def sgd(params: List[Tensor],
        params2: List[Tensor],
        d_p_list: List[Tensor],
        momentum_buffer_list: List[Optional[Tensor]],
        # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
        # setting this as kwarg for now as functional API is compiled by torch/distributed/optim
        has_sparse_grad: bool = None,
        foreach: bool = None,
        *,
        weight_decay: float,
        momentum: float,
        lr: float,
        dampening: float,
        nesterov: bool,
        maximize: bool,
        fused: bool):
    r"""Functional API that performs SGD algorithm computation.

    See :class:`~torch.optim.SGD` for details.
    """

    if foreach is None:
        # Placeholder for more complex foreach logic to be added when value is not set
        foreach = False

    if foreach and torch.jit.is_scripting():
        raise RuntimeError('torch.jit.script not supported with foreach optimizers')

    if foreach and not torch.jit.is_scripting():
        func = _multi_tensor_sgd
    else:
        func = _single_tensor_sgd

    func(params,
         params2,
         d_p_list,
         momentum_buffer_list,
         weight_decay=weight_decay,
         momentum=momentum,
         lr=lr,
         dampening=dampening,
         nesterov=nesterov,
         has_sparse_grad=has_sparse_grad,
         maximize=maximize,
         fused=fused)

@torch.no_grad()
def sgd_step(self, closure=None):
    """Performs a single optimization step.

    Args:
        closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
    """
    loss = None
    if closure is not None:
        with torch.enable_grad():
            loss = closure()

    for group in self.param_groups:
        params_with_grad = []
        params2 = []
        d_p_list = []
        momentum_buffer_list = []
        has_sparse_grad = False

        for p in group['params']:
            grad = get_bf16_grad(p, self.params_attr) if is_master_weight(p, self.params_attr) else p.grad
            if grad is not None:
                params_with_grad.append(p)
                d_p_list.append(grad)
                if grad.is_sparse:
                    has_sparse_grad = True

                state = self.state[p]
                if 'momentum_buffer' not in state:
                    momentum_buffer_list.append(None)
                else:
                    momentum_buffer_list.append(state['momentum_buffer'])

                param2 = get_param2(p, self.params_attr)
                params2.append(param2)

        sgd(params_with_grad,
            params2,
            d_p_list,
            momentum_buffer_list,
            weight_decay=group['weight_decay'],
            momentum=group['momentum'],
            lr=group['lr'],
            dampening=group['dampening'],
            nesterov=group['nesterov'],
            maximize=group['maximize'],
            has_sparse_grad=has_sparse_grad,
            foreach=group['foreach'],
            fused=self.fused)

        # update momentum_buffers in state
        for p, momentum_buffer in zip(params_with_grad, momentum_buffer_list):
            state = self.state[p]
            state['momentum_buffer'] = momentum_buffer

    return loss

def _lamb_fused_impl(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    attr: dict,
    state_steps: List[int],
    beta1: float,
    beta2: float,
    lr: float,
    weight_decay: float,
    eps: float,
):

    r"""Functional API that performs Lamb algorithm computation.
    See :class:`~torch.optim.Lamb` for details.
    """

    for i, param in enumerate(params):

        grad = grads[i]
        exp_avg = exp_avgs[i]
        exp_avg_sq = exp_avg_sqs[i]
        step = state_steps[i]
        param2 = get_param2(param, attr)
        torch.ops.torch_ipex.lamb_fused_step(
            param,
            exp_avg,
            exp_avg_sq,
            grad,
            param2,
            step,
            beta1,
            beta2,
            lr,
            weight_decay,
            eps)

def _lamb_impl(
    params: List[Tensor],
    grads: List[Tensor],
    exp_avgs: List[Tensor],
    exp_avg_sqs: List[Tensor],
    state_steps: List[int],
    beta1: float,
    beta2: float,
    lr: float,
    weight_decay: float,
    eps: float,
):
    r"""Functional API that performs Lamb algorithm computation.
    """
    for i, param in enumerate(params):
        grad = grads[i]
        exp_avg = exp_avgs[i]
        exp_avg_sq = exp_avg_sqs[i]
        step = state_steps[i]

        bias_correction1 = 1 - beta1 ** step
        bias_correction2 = 1 - beta2 ** step
        grad = grad.to(exp_avg.dtype)
        # Decay the first and second moment running average coefficient
        exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
        exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

        adam_step = (exp_avg / bias_correction1) / ((exp_avg_sq / bias_correction2).sqrt() + eps)

        if weight_decay != 0:
            adam_step.add_(param, alpha=weight_decay)

        weight_norm = param.norm(p=2)
        rtw_norm = adam_step.norm(p=2)
        true_ratio = weight_norm / rtw_norm

        param.add_(adam_step, alpha=-lr * true_ratio)

@torch.no_grad()
def lamb_step(self, closure=None):
    """Performs a single optimization step.
    Args:
        closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
    """
    loss = None
    if closure is not None:
        with torch.enable_grad():
            loss = closure()

    for group in self.param_groups:
        params_with_grad = []
        grads = []
        exp_avgs = []
        exp_avg_sqs = []
        trails = []
        state_steps = []

        for p in group['params']:
            grad = get_bf16_grad(p, self.params_attr) if is_master_weight(p, self.params_attr) else p.grad
            if grad is not None:
                params_with_grad.append(p)
                if grad.is_sparse:
                    raise RuntimeError('Lamb does not support sparse gradients')
                if grad.device != torch.device('cpu'):
                    raise RuntimeError('Lamb supports only CPU device')
                grads.append(grad)

                state = self.state[p]
                # Lazy state initialization
                if len(state) == 0:
                    state['step'] = 0
                    buffer_dtype = p.dtype if p.dtype is torch.float64 else torch.float
                    state['exp_avg'] = torch.zeros(p.shape, dtype=buffer_dtype)
                    state['exp_avg_sq'] = torch.zeros(p.shape, dtype=buffer_dtype)

                exp_avgs.append(state['exp_avg'])
                exp_avg_sqs.append(state['exp_avg_sq'])

                # update the steps for each param group update
                state['step'] += 1
                # record the step after step update
                state_steps.append(state['step'])

        beta1, beta2 = group['betas']
        _lamb_fused_impl(
            params_with_grad,
            grads,
            exp_avgs,
            exp_avg_sqs,
            self.params_attr,
            state_steps,
            beta1,
            beta2,
            group['lr'],
            group['weight_decay'],
            group['eps'])
    return loss

@torch.no_grad()
def adam_step(self, closure=None):
    """Performs a single optimization step.

    Args:
        closure (callable, optional): A closure that reevaluates the model
            and returns the loss.
    """
    loss = None
    if closure is not None:
        with torch.enable_grad():
            loss = closure()

    for group in self.param_groups:
        params_with_grad = []
        params2 = []
        grads = []
        exp_avgs = []
        exp_avg_sqs = []
        max_exp_avg_sqs = []
        state_steps = []
        beta1, beta2 = group['betas']

        for p in group['params']:
            grad = get_bf16_grad(p, self.params_attr) if is_master_weight(p, self.params_attr) else p.grad
            if grad is not None:
                params_with_grad.append(p)
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                grads.append(grad)

                state = self.state[p]
                # Lazy state initialization
                if len(state) == 0:
                    buffer_dtype = p.dtype if p.dtype is torch.float64 else torch.float
                    state['step'] = torch.tensor(0.)
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format, dtype=buffer_dtype)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format, dtype=buffer_dtype)
                    if group['amsgrad']:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format, dtype=buffer_dtype)

                exp_avgs.append(state['exp_avg'])
                exp_avg_sqs.append(state['exp_avg_sq'])

                if group['amsgrad']:
                    max_exp_avg_sqs.append(state['max_exp_avg_sq'])

                state_steps.append(state['step'])

                param2 = get_param2(p, self.params_attr)
                params2.append(param2)

        adam(params_with_grad,
                params2,
                grads,
                exp_avgs,
                exp_avg_sqs,
                max_exp_avg_sqs,
                state_steps,
                amsgrad=group['amsgrad'],
                beta1=beta1,
                beta2=beta2,
                lr=group['lr'],
                weight_decay=group['weight_decay'],
                eps=group['eps'],
                maximize=group['maximize'],
                foreach=group['foreach'])

    return loss


def adam(params: List[Tensor],
        params2: List[Tensor],
        grads: List[Tensor],
        exp_avgs: List[Tensor],
        exp_avg_sqs: List[Tensor],
        max_exp_avg_sqs: List[Tensor],
        state_steps: List[Tensor],
        # kwonly args with defaults are not supported by functions compiled with torchscript issue #70627
        # setting this as kwarg for now as functional API is compiled by torch/distributed/optim
        foreach: bool = None,
        *,
        amsgrad: bool,
        beta1: float,
        beta2: float,
        lr: float,
        weight_decay: float,
        eps: float,
        maximize: bool):
    r"""Functional API that performs Adam algorithm computation.
    See :class:`~torch.optim.Adam` for details.
    """

    if not all([isinstance(t, torch.Tensor) for t in state_steps]):
        raise RuntimeError("API has changed, `state_steps` argument must contain a list of singleton tensors")

    if foreach is None:
        # Placeholder for more complex foreach logic to be added when value is not set
        foreach = False

    if foreach and torch.jit.is_scripting():
        raise RuntimeError('torch.jit.script not supported with foreach optimizers')

    if foreach and not torch.jit.is_scripting():
        func = _multi_tensor_adam
    else:
        func = _single_tensor_adam

    func(params,
            params2,
            grads,
            exp_avgs,
            exp_avg_sqs,
            max_exp_avg_sqs,
            state_steps,
            amsgrad=amsgrad,
            beta1=beta1,
            beta2=beta2,
            lr=lr,
            weight_decay=weight_decay,
            eps=eps,
            maximize=maximize)


def _single_tensor_adam(params: List[Tensor],
                    params2: List[Tensor],
                    grads: List[Tensor],
                    exp_avgs: List[Tensor],
                    exp_avg_sqs: List[Tensor],
                    max_exp_avg_sqs: List[Tensor],
                    state_steps: List[Tensor],
                    *,
                    amsgrad: bool,
                    beta1: float,
                    beta2: float,
                    lr: float,
                    weight_decay: float,
                    eps: float,
                    maximize: bool):

    for i, param in enumerate(params):

        grad = grads[i] if not maximize else -grads[i]
        exp_avg = exp_avgs[i]
        exp_avg_sq = exp_avg_sqs[i]
        if amsgrad:
            max_exp_avg_sq = max_exp_avg_sqs[i]
        else:
            max_exp_avg_sq = torch.Tensor()
        step_t = state_steps[i]
        param2 = params2[i]
        # update step
        step_t += 1
        step = step_t.item()

        torch.ops.torch_ipex.adam_fused_step(
            param,
            exp_avg,
            exp_avg_sq,
            max_exp_avg_sq,
            grad,
            param2,
            amsgrad,
            step,
            beta1,
            beta2,
            lr,
            weight_decay,
            eps)

def _multi_tensor_adam(params: List[Tensor],
                    params2: List[Tensor],
                    grads: List[Tensor],
                    exp_avgs: List[Tensor],
                    exp_avg_sqs: List[Tensor],
                    max_exp_avg_sqs: List[Tensor],
                    state_steps: List[Tensor],
                    *,
                    amsgrad: bool,
                    beta1: float,
                    beta2: float,
                    lr: float,
                    weight_decay: float,
                    eps: float,
                    maximize: bool):

    if len(params) == 0:
        return

    if maximize:
        grads = torch._foreach_neg(tuple(grads))  # type: ignore[assignment]

    _single_tensor_adam(
        params,
        params2,
        grads,
        exp_avgs,
        exp_avg_sqs,
        max_exp_avg_sqs,
        state_steps,
        amsgrad=amsgrad,
        beta1=beta1,
        beta2=beta2,
        lr=lr,
        weight_decay=weight_decay,
        eps=eps,
        maximize=False
    )
